# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
from datetime import datetime, timedelta, timezone
from typing import List, Optional

from PyQt5.QtCore import (
    QAbstractItemModel,
    QAbstractListModel,
    QLocale,
    QModelIndex,
    QSize,
    Qt,
    QVariant,
)
from PyQt5.QtGui import QPainter, QPalette
from PyQt5.QtWidgets import QStyle, QStyledItemDelegate, QStyleOptionViewItem

from tikka.domains.application import Application
from tikka.domains.entities.identity import Certification, IdentityStatus
from tikka.interfaces.adapters.repository.certifications import (
    CertificationsRepositoryInterface,
)
from tikka.slots.pyqt.entities.constants import (
    NUMERIC_DISPLAY_COLOR_GREEN,
    NUMERIC_DISPLAY_COLOR_ORANGE,
    NUMERIC_DISPLAY_COLOR_RED,
)
from tikka.slots.pyqt.widgets.account_certification_row import (
    AccountCertificationRowWidget,
)


class CertificationItemDelegate(QStyledItemDelegate):
    """
    CertificationItemDelegate class
    """

    def __init__(self):
        """
        Init CertificationItemDelegate instance
        """
        super().__init__()

        # set item height >= AccountCertificationRowWidget.height
        self.item_height = 71

    def sizeHint(self, option: QStyleOptionViewItem, index: QModelIndex):
        """
        Return item size

        Force height

        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        return QSize(option.rect.width(), self.item_height)

    def paint(
        self, painter: QPainter, option: QStyleOptionViewItem, index: QModelIndex
    ) -> None:
        """
        Draw certification row item in listview

        :param painter: QPainter instance
        :param option: QStyleOptionViewItem instance
        :param index: QModelIndex instance
        :return:
        """
        # Init style option
        style_option = QStyleOptionViewItem(option)
        self.initStyleOption(style_option, index)

        account_address_label = index.data(
            CertificationsListModel.AccountAddressLabelRole
        )
        account_name_label = index.data(CertificationsListModel.AccountNameRole)
        identity_name_label = index.data(CertificationsListModel.IdentityNameRole)
        identity_status = index.data(CertificationsListModel.IdentityStatusRole)
        identity_status_color = index.data(
            CertificationsListModel.IdentityStatusColorRole
        )
        created_on = index.data(CertificationsListModel.CreatedOnRole)
        updated_on = index.data(CertificationsListModel.UpdatedOnRole)
        expire_on = index.data(CertificationsListModel.ExpireOnRole)
        expire_on_color = index.data(CertificationsListModel.ExpireOnColorRole)

        """ Replace item row by the widget """
        widget = AccountCertificationRowWidget(
            account_address_label,
            account_name_label,
            identity_name_label,
            identity_status,
            identity_status_color,
            created_on,
            updated_on,
            expire_on,
            expire_on_color,
        )  # Crée le widget

        # set widget background
        # widget.setAutoFillBackground(True)

        # Modify widget palette depending on behavior (hover, selection)
        palette = widget.palette()

        # 🔹 Check if the line number is even (alternate color effect)
        if index.row() % 2 == 1:  # Even lines use AlternateBase color
            base_color = option.palette.color(QPalette.AlternateBase)
        else:
            base_color = option.palette.color(QPalette.Base)

        if option.state & QStyle.State_Selected:
            palette.setColor(
                QPalette.Window, option.palette.highlight().color()
            )  # selection color
        elif option.state & QStyle.State_MouseOver:
            palette.setColor(
                QPalette.Window, option.palette.color(QPalette.Highlight).lighter(200)
            )  # Hover
        else:
            palette.setColor(
                QPalette.Window, base_color
            )  # normal or alternate background

        widget.setPalette(palette)  # Apply palette on widget
        widget.update()  # 🔥 Force display of new colors

        widget.resize(option.rect.size())  # Set size to the row size
        painter.save()
        painter.translate(option.rect.topLeft())  # Place the widget correctly
        widget.render(painter)
        painter.restore()


class CertificationsListModel(QAbstractListModel):
    """
    CertificationsListModel class that drives the population of list display
    """

    (
        AccountAddressLabelRole,
        AccountNameRole,
        IdentityNameRole,
        IdentityStatusRole,
        IdentityStatusColorRole,
        CreatedOnRole,
        UpdatedOnRole,
        ExpireOnRole,
        ExpireOnColorRole,
    ) = range(Qt.UserRole, Qt.UserRole + 9)

    def __init__(
        self,
        application: Application,
        locale: QLocale,
        filter_column: str,
    ):
        """
        Init CertificationsReceivedListModel with identity index to get identity received certification list

        :param application: Application instance
        :param locale: QLocale instance
        :param filter_column: Column of Issuer or Receiver as TransfersRepositoryInterface.COLUMN_RECEIVER_IDENTITY_INDEX
        """
        super().__init__()

        self.application = application
        self._ = self.application.translator.gettext

        # get parent widget locale to localize balance display
        self.locale = locale
        self.filter_column = filter_column
        self.identity_index: Optional[int] = None
        self.certifications: List[Certification] = []

    def init_data(
        self, identity_index: Optional[int] = None, offset: int = 0, limit: int = 1000
    ):
        """
        Fill data from repository

        :param identity_index: Identiy index or None
        :param offset: Offset in DB list to paginate
        :param limit: Limit page size
        :return:
        """
        self.beginResetModel()
        self.identity_index = identity_index
        if self.identity_index is None:
            self.certifications = []
        else:
            self.certifications = self.application.certifications.repository.list(
                self.identity_index,
                filters={self.filter_column: self.identity_index},
                offset=offset,
                limit=limit,
                sort_column=CertificationsRepositoryInterface.COLUMN_EXPIRE_ON,
                sort_order=CertificationsRepositoryInterface.SORT_ORDER_ASCENDING,
            )
            logging.debug(
                "Certifications model init data, identity #%s, column=%s, size=%s",
                self.identity_index,
                self.filter_column,
                len(self.certifications),
            )
        self.endResetModel()

    def rowCount(self, _: QModelIndex = QModelIndex()) -> int:
        """
        Return row count

        :param _: QModelIndex instance
        :return:
        """
        return len(self.certifications)

    def index(  # type: ignore
        self, row: int, column: int, parent: QModelIndex = QModelIndex()
    ) -> QModelIndex:
        """
        Return QModelIndex for row, column and parent

        :param row: Row index
        :param column: Column index
        :param parent: Parent QModelIndex instance
        :return:
        """
        if not QAbstractItemModel.hasIndex(self, row, column, parent):
            return QModelIndex()

        if row < len(self.certifications):
            return QAbstractItemModel.createIndex(
                self, row, column, self.certifications[row]
            )

        return QModelIndex()

    def data(self, index: QModelIndex, role: int = Qt.DisplayRole) -> QVariant:
        """
        Return data of cell for column index.column

        :param index: QModelIndex instance
        :param role: Item data role
        :return:
        """
        row = index.row()
        certification = self.certifications[row]
        data = QVariant()

        # set certification issued or received depending on filter column
        address = (
            certification.issuer_address
            if self.filter_column
            == self.application.repository.certifications.COLUMN_RECEIVER_IDENTITY_INDEX
            else certification.receiver_address
        )
        identity_index = (
            certification.issuer_identity_index
            if self.filter_column
            == self.application.repository.certifications.COLUMN_RECEIVER_IDENTITY_INDEX
            else certification.receiver_identity_index
        )
        identity_name = (
            certification.issuer_identity_name
            if self.filter_column
            == self.application.repository.certifications.COLUMN_RECEIVER_IDENTITY_INDEX
            else certification.receiver_identity_name
        )
        identity_status = (
            certification.issuer_identity_status
            if self.filter_column
            == self.application.repository.certifications.COLUMN_RECEIVER_IDENTITY_INDEX
            else certification.receiver_identity_status
        )

        if role == self.AccountAddressLabelRole:
            # return address
            return QVariant(f"{address[:4]}…{address[-5:]}")

        if role == self.IdentityStatusRole:
            display_identity_status = {
                IdentityStatus.UNCONFIRMED.value: self._("Unconfirmed"),
                IdentityStatus.UNVALIDATED.value: self._("Unvalidated"),
                IdentityStatus.MEMBER.value: self._("Member"),
                IdentityStatus.NOT_MEMBER.value: self._("Not member"),
                IdentityStatus.REVOKED.value: self._("Revoked"),
            }
            return QVariant(display_identity_status[identity_status.value])

        if role == self.IdentityStatusColorRole:
            display_identity_status_color = {
                IdentityStatus.UNCONFIRMED.value: "orange",
                IdentityStatus.UNVALIDATED.value: "orange",
                IdentityStatus.MEMBER.value: "green",
                IdentityStatus.NOT_MEMBER.value: "red",
                IdentityStatus.REVOKED.value: "red",
            }
            return QVariant(display_identity_status_color[identity_status.value])

        if role == self.AccountNameRole:
            account = self.application.accounts.get_by_address(address)
            if account is not None and account.name is not None:
                return QVariant(account.name)
            else:
                return QVariant()

        if role == self.IdentityNameRole:
            if identity_index is not None:
                return QVariant(f"{identity_name}#{identity_index}")
            else:
                return QVariant()

        if role == self.CreatedOnRole:
            # datetime
            datetime_label = self.locale.toString(
                certification.created_on.astimezone(),
                QLocale.dateFormat(self.locale, QLocale.ShortFormat),
            )
            return QVariant(datetime_label)

        if role == self.UpdatedOnRole:
            if certification.updated_on > certification.created_on:
                # datetime
                datetime_label = self.locale.toString(
                    certification.updated_on.astimezone(),
                    QLocale.dateFormat(self.locale, QLocale.ShortFormat),
                )
                datetime_label = self._("(renewed on {datetime})").format(
                    datetime=datetime_label
                )
            else:
                datetime_label = ""
            return QVariant(datetime_label)

        if role == self.ExpireOnRole:
            # datetime
            datetime_label = self.locale.toString(
                certification.expire_on.astimezone(),
                QLocale.dateFormat(self.locale, QLocale.LongFormat),
            )
            return QVariant(datetime_label)

        if role == self.ExpireOnColorRole:
            color = NUMERIC_DISPLAY_COLOR_GREEN
            if certification.expire_on.astimezone() < datetime.now(
                timezone.utc
            ) + timedelta(weeks=24):
                color = NUMERIC_DISPLAY_COLOR_ORANGE
            if certification.expire_on.astimezone() < datetime.now(
                timezone.utc
            ) + timedelta(weeks=8):
                color = NUMERIC_DISPLAY_COLOR_RED
            return QVariant(color)

        return data



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import List, Optional

from tikka.domains.entities.datapod import DataPod


class DataPodsRepositoryInterface(abc.ABC):
    """
    DataPodsRepositoryInterface class
    """

    COLUMN_URL = "datapod_url"
    COLUMN_BLOCK = "datapod_block"

    DEFAULT_LIST_OFFSET = 0
    DEFAULT_LIST_LIMIT = 1000

    SORT_ORDER_ASCENDING = "ASC"
    SORT_ORDER_DESCENDING = "DESC"

    @abc.abstractmethod
    def add(self, datapod: DataPod) -> None:
        """
        Add a new datapod in repository

        :param datapod: DataPod instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, url: str) -> Optional[DataPod]:
        """
        Return DataPod by url from repository

        :param url: DataPod url
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def update(self, datapod: DataPod) -> None:
        """
        Update datapod in repository

        :param datapod: DataPod instance
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(
        self,
        offset: int = DEFAULT_LIST_OFFSET,
        limit: int = DEFAULT_LIST_LIMIT,
        sort_column: str = COLUMN_URL,
        sort_order: str = SORT_ORDER_ASCENDING,
    ) -> List[DataPod]:
        """
        List datapods from repository

        :param offset: Offset index to get rows from
        :param limit: Number of rows to return
        :param sort_column: Sort column, default to DataPodsRepositoryInterface.COLUMN_URL
        :param sort_order: Sort order, default to DataPodsRepositoryInterface.ASC
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete(self, url: str) -> None:
        """
        Delete datapod by url in repository

        :param url: Url of DataPod to delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        """
        Delete all datapods in repository

        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def count(self) -> int:
        """
        Return number of datapods in repository

        :return:
        """
        raise NotImplementedError

    def get_urls(self) -> List[str]:
        """
        Get all urls of datapods from repository

        :return:
        """
        raise NotImplementedError



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from datetime import timedelta
from typing import List

from dateutil import parser
from gql import gql

from tikka.domains.entities.identity import Certification, IdentityStatus
from tikka.interfaces.adapters.network.connection import NetworkConnectionError
from tikka.interfaces.adapters.network.indexer.certifications import (
    IndexerCertificationsException,
    IndexerCertificationsInterface,
)


class IndexerCertifications(IndexerCertificationsInterface):
    """
    IndexerCertifications class
    """

    identity_status_map = {
        "Unconfirmed": IdentityStatus.UNCONFIRMED,
        "Unvalidated": IdentityStatus.UNVALIDATED,
        "Validated": IdentityStatus.MEMBER,
        "Member": IdentityStatus.MEMBER,
        "NotMember": IdentityStatus.NOT_MEMBER,
        "Revoked": IdentityStatus.REVOKED,
        "Removed": IdentityStatus.REMOVED,
    }

    def list(
        self,
        identity_index: int,
        limit: int,
        offset: int = 0,
        sort_column: str = IndexerCertificationsInterface.COLUMN_EXPIRE_ON,
        sort_order: str = IndexerCertificationsInterface.SORT_ORDER_DESCENDING,
    ) -> List[Certification]:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            IndexerCertificationsInterface.list.__doc__
        )

        if (
            not self.indexer.connection.is_connected()
            or self.indexer.connection.client is None
        ):
            raise IndexerCertificationsException(NetworkConnectionError())

        # Construire la requête GraphQL
        query = f"""
                query CertDetails {{
                blocks (
                    filter: {{ and: [
                                        {{ height: {{ greaterThanOrEqualTo: 0 }} }},
                                        {{ height: {{ lessThan: 2 }} }}
                                    ]
                            }}
                    orderBy: [HEIGHT_ASC], first: 2) {{
                    nodes {{
                      height
                      timestamp
                    }}
                  }}
                  certs(
                    first: {limit}
                    offset: {offset}
                    orderBy: {sort_column.upper()}_{sort_order.upper()}
                    filter: {{
                        or: [
                            {{ issuer: {{ index: {{ equalTo: {identity_index} }} }} }},
                            {{ receiver: {{ index: {{ equalTo: {identity_index} }} }} }}
                        ]
                    }}
                  ) {{
                    nodes {{
                          id
                          isActive
                          createdIn {{
                            block {{
                              timestamp
                            }}
                          }}
                          updatedIn {{
                            block {{
                              timestamp
                            }}
                          }}
                          expireOn
                          issuer {{
                            id
                            index
                            name
                            accountId
                            status
                          }}
                          receiver {{
                            id
                            index
                            name
                            accountId
                            status
                          }}
                    }}
                  }}
                }}
            """

        try:
            result = self.indexer.connection.client.execute(gql(query))
        except Exception as exception:
            raise IndexerCertificationsException(exception)

        # datetime are set in block since/before the block zero datetime of the blockchain
        first_two_blocks = result["blocks"]["nodes"]
        block_number_0 = first_two_blocks[0]["height"]
        assert block_number_0 == 0

        block_time_0 = parser.parse(first_two_blocks[0]["timestamp"])
        block_time_1 = parser.parse(first_two_blocks[1]["timestamp"])
        block_duration = block_time_1 - block_time_0
        avg_time_per_block = block_duration.total_seconds()

        certifications = []
        for certification in result["certs"]["nodes"]:

            expire_on_block = certification["expireOn"]

            if expire_on_block > 0:
                expire_on_time = block_time_0 + timedelta(
                    seconds=avg_time_per_block * abs(expire_on_block)
                )
            elif expire_on_block < 0:
                expire_on_time = block_time_0 - timedelta(
                    seconds=avg_time_per_block * abs(expire_on_block)
                )
            elif expire_on_block == 0:
                expire_on_time = block_time_0

            certifications.append(
                Certification(
                    id=certification["id"],
                    issuer_identity_index=certification["issuer"]["index"],
                    issuer_identity_name=certification["issuer"]["name"]
                    if certification["issuer"]["name"] != certification["issuer"]["id"]
                    else "",
                    issuer_identity_status=self.identity_status_map[
                        certification["issuer"]["status"]
                    ],
                    issuer_address=certification["issuer"]["accountId"],
                    receiver_identity_index=certification["receiver"]["index"],
                    receiver_identity_name=certification["receiver"]["name"]
                    if certification["receiver"]["name"]
                    != certification["receiver"]["id"]
                    else "",
                    receiver_identity_status=self.identity_status_map[
                        certification["receiver"]["status"]
                    ],
                    receiver_address=certification["receiver"]["accountId"],
                    created_on=parser.parse(
                        certification["createdIn"]["block"]["timestamp"]
                    ),
                    updated_on=parser.parse(
                        certification["updatedIn"]["block"]["timestamp"]
                    ),
                    expire_on=expire_on_time,
                )
            )

        return certifications



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.slots.pyqt.resources.gui.widgets.account_certification_row_rc import (
    Ui_AccountCertificationRowWidget,
)


class AccountCertificationRowWidget(QWidget, Ui_AccountCertificationRowWidget):
    """
    AccountCertificationRowWidget class
    """

    def __init__(
        self,
        account_address: str,
        account_name: str,
        identity_name: str,
        identity_status: str,
        identity_status_color: str,
        created_on: str,
        updated_on: str,
        expire_on: str,
        expire_on_color: str,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountTransferRowWidget instance

        :param account_address: Account address label text
        :param account_name: Account name
        :param identity_name: Identity name label text
        :param identity_status: Identity status
        :param identity_status_color: Identity status color
        :param created_on: Created On Datetime label text
        :param updated_on: Updated On Datetime label text
        :param expire_on: Expire On Datetime label text
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        account_identity_name_color = "blue"

        self.accountAddressValuelabel.setText(account_address)

        self.identityNameValueLabel.setStyleSheet(
            f'color: "{account_identity_name_color}"'
        )
        self.accountNameValueLabel.setText(
            f" - {account_name}" if account_name is not None else ""
        )
        self.identityNameValueLabel.setText(identity_name)
        self.identityStatusValueLabel.setText(identity_status)
        self.identityStatusValueLabel.setStyleSheet(f'color: "{identity_status_color}"')
        self.createdOnValueLabel.setText(created_on)
        self.updatedOnValueLabel.setText(updated_on)
        self.expireOnValueLabel.setText(expire_on)
        self.expireOnValueLabel.setStyleSheet(
            f"""
        color: white;
        border : 3px solid {expire_on_color};
        border-radius: 10%;
        background: {expire_on_color};"""
        )


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)

    application = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()
    account_address = "5D3r...8uI9"
    account_name = "Vince"
    identity_name = "vic#67"
    identity_status = "Member"
    identity_status_color = "green"
    created_on = "01/02/2024"
    updated_on = "(renewed on 01/02/2025)"
    expire_on = "expire on 01/02/2027"
    expire_on_color = "green"
    widget = AccountCertificationRowWidget(
        account_address,
        account_name,
        identity_name,
        identity_status,
        identity_status_color,
        created_on,
        updated_on,
        expire_on,
        expire_on_color,
        main_window,
    )
    main_window.setCentralWidget(widget)

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from __future__ import annotations

from tikka.adapters.repository.accounts import DBAccountsRepository
from tikka.adapters.repository.authorities import DBAuthoritiesRepository
from tikka.adapters.repository.categories import DBCategoriesRepository
from tikka.adapters.repository.certifications import DBCertificationsRepository
from tikka.adapters.repository.config import FileConfigRepository
from tikka.adapters.repository.currencies import FileCurrenciesRepository
from tikka.adapters.repository.currency import DBCurrencyRepository
from tikka.adapters.repository.datapods import DBDataPodsRepository
from tikka.adapters.repository.file_db_client import FileDBClient
from tikka.adapters.repository.file_wallets import V1FileWalletsRepository
from tikka.adapters.repository.identities import DBIdentitiesRepository
from tikka.adapters.repository.indexers import DBIndexersRepository
from tikka.adapters.repository.nodes import DBNodesRepository
from tikka.adapters.repository.passwords import DBPasswordsRepository
from tikka.adapters.repository.preferences import DBPreferencesRepository
from tikka.adapters.repository.profiles import DBProfilesRepository
from tikka.adapters.repository.smiths import DBSmithsRepository
from tikka.adapters.repository.technical_committee_members import (
    DBTechnicalCommitteeMembersRepository,
)
from tikka.adapters.repository.technical_committee_proposals import (
    DBTechnicalCommitteeProposalsRepository,
)
from tikka.adapters.repository.transfers import DBTransfersRepository
from tikka.adapters.repository.wallets import DBWalletsRepository
from tikka.interfaces.adapters.repository.profiles import ProfilesRepositoryInterface
from tikka.interfaces.adapters.repository.repository import RepositoryInterface
from tikka.interfaces.adapters.repository.technical_committee_proposals import (
    TechnicalCommitteeProposalsRepositoryInterface,
)


class Repository(RepositoryInterface):
    """
    Repository class
    """

    def __init__(self, data_path: str) -> None:
        """
        Init Repository instance
        """
        # dynamic user config file
        self._config = FileConfigRepository(data_path)
        # static application config file
        self._currencies = FileCurrenciesRepository(data_path)

        # handle V1 file wallets
        self._v1_file_wallets = V1FileWalletsRepository()

        # database
        self._db_client = FileDBClient()
        self._accounts = DBAccountsRepository(self._db_client)
        self._authorities = DBAuthoritiesRepository(self._db_client)
        self._categories = DBCategoriesRepository(self._db_client)
        self._currency = DBCurrencyRepository(self._db_client)
        self._identities = DBIdentitiesRepository(self._db_client)
        self._certifications = DBCertificationsRepository(self._db_client)
        self._nodes = DBNodesRepository(self._db_client)
        self._indexers = DBIndexersRepository(self._db_client)
        self._datapods = DBDataPodsRepository(self._db_client)
        self._passwords = DBPasswordsRepository(self._db_client)
        self._preferences = DBPreferencesRepository(self._db_client)
        self._smiths = DBSmithsRepository(self._db_client)
        self._transfers = DBTransfersRepository(self._db_client)
        self._wallets = DBWalletsRepository(self._db_client)
        self._technical_committee_members = DBTechnicalCommitteeMembersRepository(
            self._db_client
        )
        self._technical_committee_proposals = DBTechnicalCommitteeProposalsRepository(
            self._db_client
        )
        self._profiles = DBProfilesRepository(self._db_client)

    @property
    def db_client(self) -> FileDBClient:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.db_client.__doc__
        )
        return self._db_client

    @property
    def config(self) -> FileConfigRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.config.__doc__
        )
        return self._config

    @property
    def currencies(self) -> FileCurrenciesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.currencies.__doc__
        )
        return self._currencies

    @property
    def v1_file_wallets(self) -> V1FileWalletsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.v1_file_wallets.__doc__
        )
        return self._v1_file_wallets

    @property
    def accounts(self) -> DBAccountsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.accounts.__doc__
        )
        return self._accounts

    @property
    def authorities(self) -> DBAuthoritiesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.authorities.__doc__
        )
        return self._authorities

    @property
    def categories(self) -> DBCategoriesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.categories.__doc__
        )
        return self._categories

    @property
    def currency(self) -> DBCurrencyRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.currency.__doc__
        )
        return self._currency

    @property
    def identities(self) -> DBIdentitiesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.identities.__doc__
        )
        return self._identities

    @property
    def certifications(self) -> DBCertificationsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.certifications.__doc__
        )
        return self._certifications

    @property
    def nodes(self) -> DBNodesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.nodes.__doc__
        )
        return self._nodes

    @property
    def indexers(self) -> DBIndexersRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.indexers.__doc__
        )
        return self._indexers

    @property
    def datapods(self) -> DBDataPodsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.datapods.__doc__
        )
        return self._datapods

    @property
    def passwords(self) -> DBPasswordsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.passwords.__doc__
        )
        return self._passwords

    @property
    def preferences(self) -> DBPreferencesRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.preferences.__doc__
        )
        return self._preferences

    @property
    def smiths(self) -> DBSmithsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.smiths.__doc__
        )
        return self._smiths

    @property
    def transfers(self) -> DBTransfersRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.transfers.__doc__
        )
        return self._transfers

    @property
    def wallets(self) -> DBWalletsRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.wallets.__doc__
        )
        return self._wallets

    @property
    def technical_committee_members(
        self,
    ) -> DBTechnicalCommitteeMembersRepository:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.technical_committee_members.__doc__
        )
        return self._technical_committee_members

    @property
    def technical_committee_proposals(
        self,
    ) -> TechnicalCommitteeProposalsRepositoryInterface:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.technical_committee_proposals.__doc__
        )
        return self._technical_committee_proposals

    @property
    def profiles(
        self,
    ) -> ProfilesRepositoryInterface:
        __doc__ = (  # pylint: disable=redefined-builtin, unused-variable
            RepositoryInterface.profiles.__doc__
        )
        return self._profiles



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtGui import QMovie
from PyQt5.QtWidgets import QApplication, QListWidgetItem, QMainWindow, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.events import CurrencyEvent
from tikka.domains.entities.node import Node
from tikka.slots.pyqt.entities.constants import (
    ICON_LOADER,
    TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS,
)
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import ThreadManager
from tikka.slots.pyqt.resources.gui.widgets.technical_committee_rc import (
    Ui_technicalCommitteeWidget,
)
from tikka.slots.pyqt.widgets.technical_committee_proposal import (
    TechnicalCommitteeProposalWidget,
)


class TechnicalCommitteeWidget(QWidget, Ui_technicalCommitteeWidget):
    """
    TechnicalCommitteeWidget class
    """

    def __init__(
        self, application: Application, parent: Optional[QWidget] = None
    ) -> None:
        """
        Init TechnicalCommitteeWidget instance

        :param application: Application instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus: EventBus = EventBus.instance()
        self.event_bus.repository_data_updated.connect(
            self.on_repository_data_updated_event
        )

        self.proposals_list_current_index = 0

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        # events
        self.accountComboBox.activated.connect(self.on_account_combobox_index_changed)
        self.refreshButton.clicked.connect(self._on_refresh_button_clicked)

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            CurrencyEvent.EVENT_TYPE_CHANGED, self._on_currency_event
        )

        self.init_member_list()
        self.init_proposal_list()

    def init_member_list(self) -> None:
        """
        Init list of members

        :return:
        """
        self.membersListWidget.clear()

        for member in self.application.technical_committee.list_members():
            if member.identity_index is not None:
                label = (
                    f"{member.address} - {member.identity_name}#{member.identity_index}"
                )
            else:
                label = f"{member.address}"
            account = self.application.accounts.get_by_address(member.address)
            if account is not None and account.name is not None:
                label += f" - {account.name}"

            self.membersListWidget.addItem(label)

    def init_proposal_list(self) -> None:
        """
        Init list of proposals

        :return:
        """

        scrollbar = self.proposalsListWidget.verticalScrollBar()
        if scrollbar.maximum() > 0:
            scroll_ratio = scrollbar.value() / scrollbar.maximum()
        else:
            scroll_ratio = 0
        current_row = self.proposalsListWidget.currentRow()

        self.proposalsListWidget.clear()

        for proposal in sorted(
            self.application.technical_committee.list_proposals(),
            key=lambda proposal: proposal.voting.index,
        ):
            item = QListWidgetItem(self.proposalsListWidget)
            self.proposalsListWidget.addItem(item)
            widget = TechnicalCommitteeProposalWidget(
                self.application, proposal, parent=self
            )
            item.setSizeHint(widget.minimumSizeHint())

            # connect slot self.update_proposals to proposal item event
            widget.vote_for_proposal_start.connect(self.on_vote_for_proposal_start)
            widget.vote_for_proposal_done.connect(self.update_proposals)

            # Associate the custom widget to the list entry
            self.proposalsListWidget.setItemWidget(item, widget)

        if scrollbar.maximum() > 0:
            new_value = int(scroll_ratio * scrollbar.maximum())
            scrollbar.setValue(new_value)

        if 0 <= current_row < self.proposalsListWidget.count():
            self.proposalsListWidget.setCurrentRow(current_row)

    def init_account_combo_box(self) -> None:
        """
        Init combobox with technical committee member accounts (with wallets)

        :return:
        """
        self.accountComboBox.clear()

        accounts = self.application.accounts.get_list()
        for account in accounts:
            if account.address in [
                member.address
                for member in self.application.technical_committee.list_members()
            ] and self.application.wallets.exists(account.address):
                self.accountComboBox.addItem(
                    account.name if account.name is not None else account.address,
                    userData=account.address,
                )

        preference_account_address_selected = self.application.preferences.get(
            TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS
        )
        if preference_account_address_selected is not None:
            preference_account_selected = self.application.accounts.get_by_address(
                preference_account_address_selected
            )
            if preference_account_selected is not None:
                index = self.accountComboBox.findData(
                    preference_account_address_selected
                )
                if index > -1:
                    self.accountComboBox.setCurrentIndex(index)
                else:
                    self.application.preferences.set(
                        TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS, None
                    )

    def _on_refresh_button_clicked(self, _):
        """
        Refresh data from network when refresh button is clicked

        :param _:
        :return:
        """
        self.errorLabel.setText("")
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.thread_manager.start(
            self.fetch_members_from_network,
            self.on_fetch_members_from_network_finished,
        )
        self.thread_manager.start(
            self.fetch_proposals_from_network,
            self.on_fetch_proposals_from_network_finished,
        )

    def on_account_combobox_index_changed(self):
        """
        Triggered when account selection is changed

        :return:
        """
        address = self.accountComboBox.currentData()

        self.application.repository.preferences.set(
            TECHNICAL_COMMITTEE_SELECTED_ACCOUNT_ADDRESS,
            address,
        )

    def fetch_members_from_network(self):
        """
        Fetch members of technical committee from network

        :return:
        """
        try:
            self.application.technical_committee.network_update_members()
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def on_fetch_members_from_network_finished(self):
        """
        Triggered when async worker is finished

        :return:
        """
        logging.debug("Technical committee widget members update")

        self.init_member_list()
        self.init_account_combo_box()
        self.loader_movie.stop()
        self.loaderIconLabel.hide()

    def fetch_proposals_from_network(self):
        """
        Fetch proposals of technical committee from network

        :return:
        """
        try:
            self.application.technical_committee.network_update_proposals()
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def on_fetch_proposals_from_network_finished(self):
        """
        Triggered when async worker is finished

        :return:
        """
        logging.debug("Technical committee widget proposals update")
        self.init_proposal_list()
        self.loader_movie.stop()
        self.loaderIconLabel.hide()

    def _on_currency_event(self, _):
        """
        When a currency event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        self.thread_manager.start(
            self.fetch_members_from_network,
            self.on_fetch_members_from_network_finished,
        )
        self.thread_manager.start(
            self.fetch_proposals_from_network,
            self.on_fetch_proposals_from_network_finished,
        )

    def on_vote_for_proposal_start(self):
        """
        Vote request send, display loader waiting for vote done

        :return:
        """
        self.loader_movie.start()
        self.loaderIconLabel.show()
        self.proposals_list_current_index = (
            self.proposalsListWidget.currentIndex().row()
        )

    def update_proposals(self):
        """
        After a vote, update proposals list with new votes

        :return:
        """
        self.thread_manager.start(
            self.fetch_proposals_from_network,
            self.on_fetch_proposals_from_network_finished,
        )

    def on_repository_data_updated_event(self):
        """
        Triggered when the main window has updated repository data

        :return:
        """
        self.init_account_combo_box()
        self.init_member_list()
        self.init_proposal_list()


if __name__ == "__main__":
    logging.basicConfig(level=logging.DEBUG)
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()
    validator_node = Node(url="wss://gdev.cgeek.fr")
    if application_.nodes.get(validator_node.url) is None:
        application_.nodes.add(validator_node)
    application_.nodes.set_current_url(validator_node.url)
    application_.connections.node.connect(validator_node)
    main_window.setCentralWidget(TechnicalCommitteeWidget(application_, main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import List

from tikka.domains.entities.identity import Certification
from tikka.interfaces.adapters.network.indexer.indexer import NetworkIndexerInterface


class IndexerCertificationsInterface(abc.ABC):
    """
    IndexerCertificationsInterface class
    """

    COLUMN_EXPIRE_ON = "EXPIRE_ON"
    SORT_ORDER_ASCENDING = "ASC"
    SORT_ORDER_DESCENDING = "DESC"

    def __init__(self, indexer: NetworkIndexerInterface) -> None:
        """
        Init IndexerCertificationsInterface instance

        :param indexer: NetworkIndexerInterface instance
        :return:
        """
        self.indexer = indexer

    @abc.abstractmethod
    def list(
        self,
        identity_index: int,
        limit: int,
        offset: int = 0,
        sort_column: str = COLUMN_EXPIRE_ON,
        sort_order: str = SORT_ORDER_ASCENDING,
    ) -> List[Certification]:
        """
        Return list of certifications from and to identity_index

        :param identity_index: Identity index
        :param limit: Max number of certifications to return
        :param offset: Offset to paginate results
        :param sort_column: Sort column, default to IndexerCertificationsInterface.COLUMN_TIMESTAMP
        :param sort_order: Sort order, default to IndexerCertificationsInterface.SORT_ORDER_DESCENDING
        :return:
        """
        raise NotImplementedError


class IndexerCertificationsException(Exception):
    """
    IndexerCertificationsException class
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Optional

from tikka.domains.currencies import Currencies
from tikka.domains.entities.account import Account
from tikka.domains.entities.profile import Profile
from tikka.interfaces.adapters.network.datapod.profiles import DataPodProfilesInterface
from tikka.interfaces.adapters.repository.profiles import ProfilesRepositoryInterface
from tikka.libs.keypair import Keypair


class Profiles:

    """
    Profiles domain class
    """

    def __init__(
        self,
        repository: ProfilesRepositoryInterface,
        datapod_profiles: DataPodProfilesInterface,
        currencies: Currencies,
    ):
        """
        Init Profiles domain

        :param repository: ProfilesRepositoryInterface instance
        :param datapod_profiles: DataPodProfilesInterface instance
        :param currencies: Currencies domain instance
        """
        self.repository = repository
        self.datapod_profiles = datapod_profiles
        self.currencies = currencies

    def add(self, profile: Profile) -> None:
        """
        Add a new profile in repository

        :param profile: Profile instance
        :return:
        """
        self.repository.add(profile)

    def get(self, address: str) -> Optional[Profile]:
        """
        Return Profile instance from repository

        :param address: Profile account address
        :return:
        """
        return self.repository.get(address)

    def update(self, profile: Profile) -> None:
        """
        Update profile in repository

        :param profile: Profile instance
        :return:
        """
        self.repository.update(profile)

    def delete(self, address: str) -> None:
        """
        Delete profile in repository

        :param address: Profile account address to delete
        :return:
        """
        self.repository.delete(address)

    def delete_all(self) -> None:
        """
        Delete all profiles in repository

        :return:
        """
        self.repository.delete_all()

    def exists(self, address: str) -> bool:
        """
        Return True if profile exists in repository

        :param address: Profile account address to check
        :return:
        """
        return self.repository.exists(address)

    def network_update(self, account: Account) -> Optional[Profile]:
        """
        Return Profile instance by account address from network if any

        :param account: Profile account address
        :return:
        """
        profile = self.datapod_profiles.get(account.address)
        if profile is not None:
            # # fixme: We get profiles from v1 datapod for now.
            # #        But we store them with the v2 address as primary key
            # profile.address = account.address
            if self.get(account.address) is not None:
                self.update(profile)
            else:
                self.add(profile)
        elif self.get(account.address) is not None:
            self.delete(account.address)

        return profile

    def network_add(self, keypair: Keypair, profile: Profile) -> None:
        """
        Add Profile to network signed by keypair

        :param keypair: Owner keypair for issuer address and signature
        :param profile: Profile for data
        :return:
        """
        self.datapod_profiles.add(keypair, profile)

    def network_delete(self, keypair: Keypair) -> None:
        """
        Delete Profile from network

        :param keypair: Owner keypair for profile address and signature
        :return:
        """
        self.datapod_profiles.delete(keypair)



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtGui import QFont, QMovie
from PyQt5.QtWidgets import QApplication, QDialog, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account
from tikka.domains.entities.constants import DATA_PATH
from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.slots.pyqt.entities.constants import ADDRESS_MONOSPACE_FONT_NAME, ICON_LOADER
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import (
    ThreadManager,
)
from tikka.slots.pyqt.resources.gui.windows.identity_change_account_rc import (
    Ui_IdentityChangeAccountDialog,
)
from tikka.slots.pyqt.windows.account_unlock import AccountUnlockWindow


class IdentityChangeAccountWindow(QDialog, Ui_IdentityChangeAccountDialog):
    """
    IdentityChangeAccountWindow class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        parent: Optional[QWidget] = None,
    ):
        """
        Init identity change account window

        :param application: Application instance
        :param account: Account owner instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus = EventBus.instance()

        self.account = account
        self.destination_account: Optional[Account] = None
        self.destination_identity: Optional[Identity] = None

        # animated loading icon
        self.loader_movie = QMovie(ICON_LOADER)
        self.loader_movie.start()
        self.loaderIconLabel.setMovie(self.loader_movie)
        loader_icon_size_policy = self.loaderIconLabel.sizePolicy()
        loader_icon_size_policy.setRetainSizeWhenHidden(True)
        self.loaderIconLabel.setSizePolicy(loader_icon_size_policy)
        self.loaderIconLabel.hide()

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(self.monospace_font)

        self.changeAccountButton.setEnabled(False)

        # events
        self.accountComboBox.activated.connect(self.on_account_combobox_index_changed)
        self.changeAccountButton.clicked.connect(self._on_change_account_button_clicked)

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        self.init_account_combo_box()

    def init_account_combo_box(self) -> None:
        """
        Init combobox with no identity accounts
        Avec séparateur entre noms et adresses
        """
        self.accountComboBox.clear()
        self.accountComboBox.addItem("-", userData=None)

        # Séparer les comptes
        named_accounts = []
        unnamed_accounts = []

        for account in self.application.accounts.get_list():
            identity = self.application.identities.get_index_by_address(account.address)
            if identity is None:
                if account.name and account.name.strip():
                    named_accounts.append(account)
                else:
                    unnamed_accounts.append(account)

        # Trier
        named_accounts.sort(key=lambda acc: acc.name.lower())  # type: ignore
        unnamed_accounts.sort(key=lambda acc: acc.address.lower())

        # Ajouter les comptes avec nom
        if named_accounts:
            for account in named_accounts:
                self.accountComboBox.addItem(
                    f"{account.address} - {account.name}",
                    userData=account.address,
                )

            # Ajouter un séparateur si on a les deux catégories
            if unnamed_accounts:
                self.accountComboBox.insertSeparator(len(named_accounts) + 1)

        # Ajouter les comptes sans nom
        for account in unnamed_accounts:
            self.accountComboBox.addItem(
                account.address,
                userData=account.address,
            )

    def on_account_combobox_index_changed(self):
        """
        Triggered when account selection is changed

        :return:
        """
        destination_address = self.accountComboBox.currentData()
        self.destination_account = self.application.accounts.get_by_address(
            destination_address
        )
        self.update_ui()

    def update_ui(self):
        """
        Update UI fields

        :return:
        """
        self.addressValueLabel.setText(self.destination_account.address or "")
        self.changeAccountButton.setEnabled(
            self.accountComboBox.currentData() is not None
        )

        if self.destination_identity is not None:
            display_identity_status = {
                IdentityStatus.UNCONFIRMED.value: self._("Unconfirmed"),
                IdentityStatus.UNVALIDATED.value: self._("Unvalidated"),
                IdentityStatus.MEMBER.value: self._("Member"),
                IdentityStatus.NOT_MEMBER.value: self._("Not member"),
                IdentityStatus.REVOKED.value: self._("Revoked"),
            }
            identity_status_color = {
                IdentityStatus.UNCONFIRMED.value: "orange",
                IdentityStatus.UNVALIDATED.value: "orange",
                IdentityStatus.MEMBER.value: "green",
                IdentityStatus.NOT_MEMBER.value: "red",
                IdentityStatus.REVOKED.value: "red",
            }
            identity_status_style = f"""
                                                    color: white;
                                                    border: 3px solid "{identity_status_color[self.destination_identity.status.value]}";
                                                    border-radius: 10%;
                                                    background: "{identity_status_color[self.destination_identity.status.value]}";
                                                    """

            self.identityIndexValueLabel.setText(f"#{self.destination_identity.index}")
            self.identityStatusValueLabel.setText(
                display_identity_status[self.destination_identity.status.value]
            )
            self.identityStatusValueLabel.setStyleSheet(identity_status_style)
        else:
            self.identityIndexValueLabel.setText("")
            self.identityStatusValueLabel.setText("")
            self.identityStatusValueLabel.setStyleSheet("")

    def _on_change_account_button_clicked(self):
        """
        Run when use click change account button

        :return:
        """
        if self.destination_account is None:
            self.errorLabel.setStyleSheet("color: red;")
            self.errorLabel.setText(self._("No account selected"))
            return

        if not self.application.wallets.is_unlocked(self.account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        if not self.application.wallets.is_unlocked(self.destination_account.address):
            # ask password...
            dialog_code = AccountUnlockWindow(
                self.application, self.destination_account, self
            ).exec_()
            if dialog_code == QDialog.Rejected:
                return

        self.loaderIconLabel.show()
        self.changeAccountButton.setDisabled(True)
        self.thread_manager.start(
            self.network_change_account, self._on_finished_network_change_account
        )

    def network_change_account(self):
        """
        Change owner account for identity on the network

        :return:
        """
        if self.destination_account is not None:
            try:
                self.application.identities.network_change_owner_key(
                    self.application.wallets.get_keypair(self.account.address),
                    self.application.wallets.get_keypair(
                        self.destination_account.address
                    ),
                )
            except Exception as exception:
                self.errorLabel.setStyleSheet("color: red;")
                self.errorLabel.setText(str(exception))
                logging.exception(exception)

            try:
                self.application.identities.network_update_identity(
                    self.account.address
                )
            except Exception as exception:
                logging.exception(exception)

            try:
                self.application.identities.network_update_identity(
                    self.destination_account.address
                )
            except Exception as exception:
                logging.exception(exception)

    def _on_finished_network_change_account(self):
        """
        When async network_change_account is finished

        :return:
        """
        self.destination_identity = self.application.identities.get_by_address(
            self.destination_account.address
        )
        self.loaderIconLabel.hide()
        self.update_ui()
        if self.destination_identity is not None:
            self.errorLabel.setStyleSheet("color: green;")
            self.errorLabel.setText(
                self._("Identity #{index} moved to {account_address}").format(
                    index=self.destination_identity.index,
                    account_address=self.destination_account.address,
                )
            )
            self.init_account_combo_box()
            self.changeAccountButton.setEnabled(False)
            self.event_bus.identity_account_changed.emit(
                self.account, self.destination_account
            )
            logging.debug(
                "identity_account_changed_signal emitted from identity_change_account window %s -> #%s",
                self.account.address,
                self.destination_account.address,
            )
        else:
            self.changeAccountButton.setEnabled(True)


if __name__ == "__main__":
    qapp = QApplication(sys.argv)
    application_ = Application(DATA_PATH)
    account_ = Account("xxxx")
    IdentityChangeAccountWindow(application_, account_).exec_()



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from typing import Optional

from PyQt5.QtCore import QPoint, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QApplication, QMainWindow, QStyledItemDelegate, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.constants import (
    DATA_PATH,
    DATAPOD_CESIUM_PLUS_V1_BLOCK_ZERO_HASH,
)
from tikka.domains.entities.datapod import DataPod
from tikka.domains.entities.events import (
    DataPodsEvent,
    IndexersEvent,
    NodesEvent,
)
from tikka.domains.entities.indexer import Indexer
from tikka.domains.entities.node import Node
from tikka.slots.pyqt.entities.constants import (
    DATAPODS_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    DATAPODS_TABLE_SORT_ORDER_PREFERENCES_KEY,
    INDEXERS_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    INDEXERS_TABLE_SORT_ORDER_PREFERENCES_KEY,
    NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY,
    NODES_TABLE_SORT_ORDER_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import ThreadManager
from tikka.slots.pyqt.models.datapods import DataPodsTableModel
from tikka.slots.pyqt.models.indexers import IndexersTableModel
from tikka.slots.pyqt.models.nodes import NodesTableModel
from tikka.slots.pyqt.resources.gui.widgets.servers_rc import Ui_ServersWidget
from tikka.slots.pyqt.widgets.datapod_menu import DataPodPopupMenu
from tikka.slots.pyqt.widgets.indexer_menu import IndexerPopupMenu
from tikka.slots.pyqt.widgets.node_menu import NodePopupMenu
from tikka.slots.pyqt.windows.datapod_add import DataPodAddWindow
from tikka.slots.pyqt.windows.indexer_add import IndexerAddWindow
from tikka.slots.pyqt.windows.node_add import NodeAddWindow


class HighlightNodeDelegate(QStyledItemDelegate):
    """
    Class to highlight connected node row
    """

    def __init__(self, table_model, application):
        """
        Initialize HighlightNodeDelegate instance

        :param table_model: TableModel of
        :param application:
        """
        super().__init__()
        self.table_model = table_model
        self.application = application

    def paint(self, painter, option, index):
        # get current url/row node
        current_node_url = self.application.nodes.get_current_url()
        node = self.table_model.nodes[index.row()]
        # check connection...
        if node.url == current_node_url:
            color = (
                QColor("lightgreen")
                if self.application.connections.node.is_connected()
                else QColor("lightcoral")
            )
            painter.fillRect(option.rect, color)

        # paint rectangle
        super().paint(painter, option, index)


class HighlightIndexerDelegate(QStyledItemDelegate):
    """
    Class to highlight connected indexer row
    """

    def __init__(self, table_model, application):
        """
        Initialize HighlightIndexerDelegate instance

        :param table_model: TableModel of
        :param application:
        """
        super().__init__()
        self.table_model = table_model
        self.application = application

    def paint(self, painter, option, index):
        # get current url/row indexer
        current_indexer_url = self.application.indexers.get_current_url()
        indexer = self.table_model.indexers[index.row()]
        # check connection...
        if indexer.url == current_indexer_url:
            color = (
                QColor("lightgreen")
                if self.application.connections.indexer.is_connected()
                else QColor("lightcoral")
            )
            painter.fillRect(option.rect, color)

        # paint rectangle
        super().paint(painter, option, index)


class HighlightDataPodDelegate(QStyledItemDelegate):
    """
    Class to highlight connected datapod row
    """

    def __init__(self, table_model, application):
        """
        Initialize HighlightDataPodDelegate instance

        :param table_model: TableModel of
        :param application:
        """
        super().__init__()
        self.table_model = table_model
        self.application = application

    def paint(self, painter, option, index):
        # get current url/row indexer
        current_datapod_url = self.application.datapods.get_current_url()
        datapod = self.table_model.datapods[index.row()]
        # check connection...
        if datapod.url == current_datapod_url:
            color = (
                QColor("lightgreen")
                if self.application.connections.datapod.is_connected()
                else QColor("lightcoral")
            )
            painter.fillRect(option.rect, color)

        # paint rectangle
        super().paint(painter, option, index)


class ServersWidget(QWidget, Ui_ServersWidget):
    """
    ServersWidget class
    """

    def __init__(
        self,
        application: Application,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init ServersWidget instance

        :param application: Application instance
        :param parent: MainWindow instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus = EventBus.instance()
        self.event_bus.node_connected.connect(self.on_node_connection_event)
        self.event_bus.node_disconnected.connect(self.on_node_connection_event)
        self.event_bus.indexer_connected.connect(self.on_indexer_connection_event)
        self.event_bus.indexer_disconnected.connect(self.on_indexer_connection_event)
        self.event_bus.datapod_connected.connect(self.on_datapod_connection_event)
        self.event_bus.datapod_disconnected.connect(self.on_datapod_connection_event)
        self.event_bus.currency_opened.connect(self.on_currency_opened_event)

        self.selected_node: Optional[Node] = None
        self.selected_indexer: Optional[Indexer] = None
        self.selected_datapod: Optional[DataPod] = None

        self.nodes_table_model = NodesTableModel(self.application)
        self.nodesTableView.setModel(self.nodes_table_model)
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()
        self.nodesTableView.customContextMenuRequested.connect(
            self.on_node_context_menu
        )
        self.nodesTableView.setItemDelegate(
            HighlightNodeDelegate(self.nodes_table_model, self.application)
        )
        pref_sort_column = self.application.repository.preferences.get(
            NODES_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if pref_sort_column is None:
            sort_column = 0
        else:
            sort_column = int(pref_sort_column)

        pref_sort_order = self.application.repository.preferences.get(
            NODES_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        if pref_sort_order is None:
            sort_order = int(Qt.SortOrder.AscendingOrder)
        else:
            sort_order = int(pref_sort_order)
        self.nodesTableView.sortByColumn(sort_column, sort_order)

        self.indexers_table_model = IndexersTableModel(self.application)
        self.indexersTableView.setModel(self.indexers_table_model)
        self.indexersTableView.resizeColumnsToContents()
        self.indexersTableView.resizeRowsToContents()
        self.indexersTableView.customContextMenuRequested.connect(
            self.on_indexer_context_menu
        )
        self.indexersTableView.setItemDelegate(
            HighlightIndexerDelegate(self.indexers_table_model, self.application)
        )
        pref_sort_column = self.application.repository.preferences.get(
            INDEXERS_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if pref_sort_column is None:
            sort_column = 0
        else:
            sort_column = int(pref_sort_column)

        pref_sort_order = self.application.repository.preferences.get(
            INDEXERS_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        if pref_sort_order is None:
            sort_order = int(Qt.SortOrder.AscendingOrder)
        else:
            sort_order = int(pref_sort_order)
        self.indexersTableView.sortByColumn(sort_column, sort_order)

        self.datapods_table_model = DataPodsTableModel(self.application)
        self.dataPodsTableView.setModel(self.datapods_table_model)
        self.dataPodsTableView.resizeColumnsToContents()
        self.dataPodsTableView.resizeRowsToContents()
        self.dataPodsTableView.customContextMenuRequested.connect(
            self.on_datapod_context_menu
        )
        self.dataPodsTableView.setItemDelegate(
            HighlightDataPodDelegate(self.datapods_table_model, self.application)
        )
        pref_sort_column = self.application.repository.preferences.get(
            DATAPODS_TABLE_SORT_COLUMN_PREFERENCES_KEY
        )
        if pref_sort_column is None:
            sort_column = 0
        else:
            sort_column = int(pref_sort_column)

        pref_sort_order = self.application.repository.preferences.get(
            DATAPODS_TABLE_SORT_ORDER_PREFERENCES_KEY
        )
        if pref_sort_order is None:
            sort_order = int(Qt.SortOrder.AscendingOrder)
        else:
            sort_order = int(pref_sort_order)
        self.dataPodsTableView.sortByColumn(sort_column, sort_order)

        # events
        self.addNodeButton.clicked.connect(self._on_add_node_button_clicked)
        self.addIndexerButton.clicked.connect(self._on_add_indexer_button_clicked)
        self.addDataPodButton.clicked.connect(self._on_add_datapod_button_clicked)
        self.nodesTableView.doubleClicked.connect(self.on_node_table_double_click)
        self.indexersTableView.doubleClicked.connect(self.on_indexer_table_double_click)
        self.dataPodsTableView.doubleClicked.connect(self.on_datapod_table_double_click)

        # subscribe to application events
        self.application.event_dispatcher.add_event_listener(
            NodesEvent.EVENT_TYPE_LIST_CHANGED, self._on_node_list_changed_event
        )
        self.application.event_dispatcher.add_event_listener(
            IndexersEvent.EVENT_TYPE_LIST_CHANGED, self._on_indexer_list_changed_event
        )
        self.application.event_dispatcher.add_event_listener(
            DataPodsEvent.EVENT_TYPE_LIST_CHANGED, self._on_datapod_list_changed_event
        )

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

    def on_node_connection_event(self):
        """
        Triggered when node connection has changed

        :return:
        """
        self.nodesTableView.viewport().update()

    def on_indexer_connection_event(self):
        """
        Triggered when indexer connection has changed

        :return:
        """
        self.indexersTableView.viewport().update()

    def on_datapod_connection_event(self):
        """
        Triggered when datapod connection has changed

        :return:
        """
        self.dataPodsTableView.viewport().update()

    def _on_add_node_button_clicked(self):
        """
        Trigger when user click on add node button

        :return:
        """
        NodeAddWindow(self.application, self).exec_()

    def _on_add_indexer_button_clicked(self):
        """
        Trigger when user click on add indexer button

        :return:
        """
        IndexerAddWindow(self.application, self).exec_()

    def _on_add_datapod_button_clicked(self):
        """
        Trigger when user click on add data pod button

        :return:
        """
        DataPodAddWindow(self.application, self).exec_()

    def on_currency_opened_event(self, _):
        """
        When a currency opened event is triggered

        :param _: CurrencyEvent instance
        :return:
        """
        # update model
        self.nodesTableView.model().init_data()
        self.indexersTableView.model().init_data()
        self.dataPodsTableView.model().init_data()

        # resize view
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()
        self.indexersTableView.resizeColumnsToContents()
        self.indexersTableView.resizeRowsToContents()
        self.dataPodsTableView.resizeColumnsToContents()
        self.dataPodsTableView.resizeRowsToContents()

    def _on_node_list_changed_event(self, _):
        """
        When the node list has changed

        :param _: NodesEvent instance
        :return:
        """
        # update model
        self.nodesTableView.model().init_data()
        # resize view
        self.nodesTableView.resizeColumnsToContents()
        self.nodesTableView.resizeRowsToContents()

    def _on_indexer_list_changed_event(self, _):
        """
        When the indexer list has changed

        :param _: IndexersEvent instance
        :return:
        """
        # update model
        self.indexersTableView.model().init_data()
        # resize view
        self.indexersTableView.resizeColumnsToContents()
        self.indexersTableView.resizeRowsToContents()

    def _on_datapod_list_changed_event(self, _):
        """
        When the data pod list has changed

        :param _: DataPodsEvent instance
        :return:
        """
        # update model
        self.dataPodsTableView.model().init_data()
        # resize view
        self.dataPodsTableView.resizeColumnsToContents()
        self.dataPodsTableView.resizeRowsToContents()

    def on_node_context_menu(self, position: QPoint):
        """
        When right button on node table view

        :param position: QPoint instance
        :return:
        """
        index = self.nodesTableView.indexAt(position)
        if index.isValid():
            # get selected node
            row = index.row()
            node = self.nodes_table_model.nodes[row]
            # display popup menu at click position
            NodePopupMenu(self.application, node).exec_(
                self.nodesTableView.mapToGlobal(position)
            )

    def on_indexer_context_menu(self, position: QPoint):
        """
        When right button on indexer table view

        :param position: QPoint instance
        :return:
        """
        index = self.indexersTableView.indexAt(position)
        if index.isValid():
            # get selected node
            row = index.row()
            indexer = self.indexers_table_model.indexers[row]
            # display popup menu at click position
            IndexerPopupMenu(self.application, indexer).exec_(
                self.indexersTableView.mapToGlobal(position)
            )

    def on_datapod_context_menu(self, position: QPoint):
        """
        When right button on data pod table view

        :param position: QPoint instance
        :return:
        """
        index = self.dataPodsTableView.indexAt(position)
        if index.isValid():
            # get selected node
            row = index.row()
            datapod = self.datapods_table_model.datapods[row]
            # display popup menu at click position
            DataPodPopupMenu(self.application, datapod).exec_(
                self.dataPodsTableView.mapToGlobal(position)
            )

    def on_node_table_double_click(self, index):
        """
        Triggered when user double-click on a row of node table view
        """
        row = index.row()
        # check index...
        if 0 <= row < len(self.nodes_table_model.nodes):
            self.selected_node = self.nodes_table_model.nodes[row]
            self.thread_manager.start(
                self.toggle_node_connection, self._on_finished_toggle_node_connection
            )

    def toggle_node_connection(self):
        """
        Toggle node connection

        :return:
        """
        if self.selected_node is None:
            return
        # current url connected
        current_node_url = self.application.nodes.get_current_url()

        # check connection...
        if (
            self.selected_node.url == current_node_url
            and self.application.connections.node.is_connected()
        ):
            self.application.connections.node.disconnect()
        else:
            if self.selected_node.url != current_node_url:
                self.application.nodes.set_current_url(self.selected_node.url)

            self.application.connections.node.connect(self.selected_node)
            if self.application.connections.node.is_connected():
                node_currency = (
                    self.application.currencies.network.currency.get_instance()
                )
                if not self.application.nodes.check_genesis_hash(
                    node_currency.genesis_hash
                ):
                    self.application.connections.node.disconnect()
                    logging.error(
                        "Node currency hash %s is different! Force Disconnect!",
                        node_currency.genesis_hash,
                    )

    def _on_finished_toggle_node_connection(self):
        """
        Triggered when toggle_node_connection is finished

        :return:
        """
        if self.application.connections.node.is_connected():
            self.event_bus.node_connected.emit()
        else:
            self.event_bus.node_disconnected.emit()

    def on_indexer_table_double_click(self, index):
        """
        Triggered when user double-click on a row of indexer table view
        """
        row = index.row()
        # check index...
        if 0 <= row < len(self.indexers_table_model.indexers):
            self.selected_indexer = self.indexers_table_model.indexers[row]
            self.thread_manager.start(
                self.toggle_indexer_connection,
                self._on_finished_toggle_indexer_connection,
            )

    def toggle_indexer_connection(self):
        """
        Toggle indexer connection

        :return:
        """
        if self.selected_indexer is None:
            return
        # current url connected
        current_indexer_url = self.application.indexers.get_current_url()

        # check connection...
        if (
            self.selected_indexer.url == current_indexer_url
            and self.application.connections.indexer.is_connected()
        ):
            self.application.connections.indexer.disconnect()
        else:
            if self.selected_indexer.url != current_indexer_url:
                self.application.indexers.set_current_url(self.selected_indexer.url)

            self.application.connections.indexer.connect(self.selected_indexer)
            if self.application.connections.indexer.is_connected():
                indexer_genesis_hash = (
                    self.application.indexers.network_get_genesis_hash()
                )
                if not self.application.indexers.check_genesis_hash(
                    indexer_genesis_hash
                ):
                    self.application.connections.indexer.disconnect()
                    logging.error("Indexer currency is different! Force Disconnect!")

    def _on_finished_toggle_indexer_connection(self):
        """
        Triggered when toggle_indexer_connection is finished

        :return:
        """
        if self.application.connections.indexer.is_connected():
            self.event_bus.indexer_connected.emit()
        else:
            self.event_bus.indexer_disconnected.emit()

    def on_datapod_table_double_click(self, index):
        """
        Triggered when user double-click on a row of datapod table view
        """
        row = index.row()
        # check index...
        if 0 <= row < len(self.datapods_table_model.datapods):
            self.selected_datapod = self.datapods_table_model.datapods[row]
            self.thread_manager.start(
                self.toggle_datapod_connection,
                self._on_finished_toggle_datapod_connection,
            )

    def toggle_datapod_connection(self):
        """
        Toggle data pod connection

        :return:
        """
        if self.selected_datapod is None:
            return
        # current url connected
        current_datapod_url = self.application.datapods.get_current_url()

        # check connection...
        if (
            self.selected_datapod.url == current_datapod_url
            and self.application.connections.datapod.is_connected()
        ):
            self.application.connections.datapod.disconnect()
        else:
            if self.selected_datapod.url != current_datapod_url:
                self.application.datapods.set_current_url(self.selected_datapod.url)

            self.application.connections.datapod.connect(self.selected_datapod)
            if self.application.connections.datapod.is_connected():
                datapod_genesis_hash = (
                    self.application.datapods.network_get_genesis_hash()
                )
                if (
                    "localhost" not in self.application.connections.datapod.url
                    and "127.0.0.1" not in self.application.connections.datapod.url
                    and datapod_genesis_hash != DATAPOD_CESIUM_PLUS_V1_BLOCK_ZERO_HASH
                ):
                    self.application.connections.datapod.disconnect()
                    logging.error("Data Pod currency is different! Force Disconnect!")

    def _on_finished_toggle_datapod_connection(self):
        """
        Triggered when toggle_datapod_connection is finished

        :return:
        """
        if self.application.connections.datapod.is_connected():
            self.event_bus.datapod_connected.emit()
        else:
            self.event_bus.datapod_disconnected.emit()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(ServersWidget(application_, main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from pathlib import Path
from time import sleep
from typing import List, Optional

from PyQt5.QtCore import QCoreApplication, QLibraryInfo, QLocale, QTranslator
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication

from tikka.domains.application import Application
from tikka.domains.config import Config
from tikka.domains.entities.account import Account
from tikka.domains.entities.identity import Identity
from tikka.slots.pyqt.entities.constants import ACCOUNT_TRANSFER_HISTORY_LIMIT
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.worker import ThreadManager
from tikka.slots.pyqt.widgets.account import AccountWidget
from tikka.slots.pyqt.windows.main import MainWindow


class PyQtMainApplication:
    """
    PyQtMainApplication class
    """

    def __init__(
        self, application: Application
    ):  # pylint: disable=super-init-not-called
        """
        Init PyQtMainApplication instance

        :param application: Application instance
        """

        self.application = application
        self.main_window: Optional[MainWindow] = None

        # Async manager
        self.thread_manager: ThreadManager = ThreadManager()

        # subscribe to the event bus
        self.event_bus = EventBus.instance()
        self.event_bus.main_window_shown.connect(self.on_window_shown)
        self.event_bus.user_data_restored.connect(self.on_user_data_restored_event)
        self.event_bus.currency_changed.connect(self.on_currency_changed)
        self.event_bus.transfer_sent.connect(
            lambda _accounts: self.thread_manager.start(
                lambda: self.on_transfer_sent(_accounts),
                lambda: self.on_transfer_sent_finished(_accounts),
            )
        )
        self.event_bus.identity_created.connect(
            lambda issuer_identity, identity: self.thread_manager.start(
                lambda: self.on_identity_created(issuer_identity, identity),
                lambda: self.on_identity_created_finished(issuer_identity, identity),
            )
        )
        self.event_bus.identity_confirmed.connect(
            lambda identity: self.thread_manager.start(
                lambda: self.on_identity_confirmed(identity),
                lambda: self.on_identity_confirmed_finished(identity),
            )
        )
        self.event_bus.identity_certified.connect(
            lambda issuer_identity, receiver_identity: self.thread_manager.start(
                lambda: self.on_identity_certified(issuer_identity, receiver_identity),
                lambda: self.on_identity_certified_finished(
                    issuer_identity, receiver_identity
                ),
            )
        )
        self.event_bus.identity_account_changed.connect(
            lambda issuer_account, destination_account: self.thread_manager.start(
                lambda: self.on_identity_account_changed(
                    issuer_account, destination_account
                ),
                lambda: self.on_identity_account_changed_finished(
                    issuer_account, destination_account
                ),
            )
        )
        self.event_bus.profile_account_changed.connect(
            lambda issuer_account, destination_account: self.thread_manager.start(
                lambda: self.on_profile_account_changed(
                    issuer_account, destination_account
                ),
                lambda: self.on_profile_account_changed_finished(
                    issuer_account, destination_account
                ),
            )
        )

    def run(self):
        """
        Start PyQt application event loop

        :return:
        """
        qt_application = QApplication(sys.argv)

        self.init_i18n(qt_application)

        # main window
        qt_application.setWindowIcon(QIcon(":/icons/logo"))
        self.main_window = MainWindow(self.application)
        self.main_window.show()

        # open welcome window if no accounts
        if self.application.accounts.count() == 0:
            self.main_window.open_welcome_window()

        # run Qt event loop
        sys.exit(qt_application.exec_())

    def init_i18n(self, qt_application: QApplication):
        """
        Init Locale and translator

        :param qt_application: QApplication instance
        :return:
        """
        # qt builtin locales for filedialog or buttonbox...
        QLocale.setDefault(QLocale(self.application.config.get(Config.LANGUAGE_KEY)))
        translator = QTranslator(qt_application)
        filename = (
            "qtbase_" + self.application.config.get(Config.LANGUAGE_KEY).split("_")[0]
        )
        result = translator.load(
            filename, QLibraryInfo.location(QLibraryInfo.TranslationsPath)
        )
        if result is False:
            raise FileNotFoundError(
                Path(QLibraryInfo.location(QLibraryInfo.TranslationsPath)).joinpath(
                    filename
                )
            )
        QCoreApplication.installTranslator(translator)

    def connect_to_servers_and_fetch_data(self):
        """
        Connect to node, indexer and datapod and fetch data from network
        to update local repository
        :return:
        """
        self.connect_to_node()
        self.connect_to_indexer()
        self.connect_to_datapod()

    def connect_to_node(self):
        """
        Connection to node

        :return:
        """
        self.application.nodes.connect()
        if self.application.connections.node.is_connected():
            self.event_bus.node_connected.emit()
            self.fetch_data_from_network_node()
        else:
            self.event_bus.node_disconnected.emit()

    def connect_to_indexer(self):
        """
        Connection to indexer

        :return:
        """
        self.application.indexers.connect()
        if self.application.connections.indexer.is_connected():
            self.event_bus.indexer_connected.emit()
            self.fetch_data_from_network_indexer()
        else:
            self.event_bus.indexer_disconnected.emit()

    def connect_to_datapod(self):
        """
        Connection to datapod

        :return:
        """
        self.application.datapods.connect()
        if self.application.connections.datapod.is_connected():
            self.event_bus.datapod_connected.emit()
            self.fetch_data_from_network_datapod()
        else:
            self.event_bus.datapod_disconnected.emit()

    def fetch_data_from_network_node(self):
        """
        Fetch data from the network node

        :return:
        """
        self.application.nodes.network_fetch_current_node()
        self.application.currencies.network_update_properties()

        self.application.nodes.network_fetch_endpoints()
        self.application.indexers.network_fetch_endpoints()

        accounts = self.application.accounts.get_list()
        try:
            self.application.accounts.network_update_balances(accounts)
        except Exception as exception:
            logging.exception(exception)

        addresses = [account.address for account in accounts]
        try:
            self.application.identities.network_update_identities(addresses)
        except Exception as exception:
            logging.exception(exception)
        else:
            identity_indice = self.application.identities.list_indice()
            try:
                self.application.smiths.network_update_smiths(identity_indice)
            except Exception as exception:
                logging.exception(exception)
            else:
                self.application.authorities.network_get_all()

        self.application.technical_committee.network_update_members()
        self.application.technical_committee.network_update_proposals()

    def fetch_data_from_network_indexer(self):
        """
        Fetch data from the network indexer

        :return:
        """
        self.application.indexers.network_fetch_current_indexer()

        accounts = self.application.accounts.get_list()
        addresses = [account.address for account in accounts]
        try:
            # update identity names from indexer
            self.application.identities.network_update_identities(addresses)
        except Exception as exception:
            logging.exception(exception)

        for identity in self.application.identities.list():
            self.application.certifications.network_fetch_certifications_for_identity(
                identity.index
            )

        # update identity names for technical committee members from indexer
        self.application.technical_committee.network_update_members()

        # update transfers for all account opened in main window
        for widget in self.main_window.get_tab_widgets_by_class(AccountWidget):
            if isinstance(widget, AccountWidget):
                self.application.transfers.network_fetch_history_for_account(
                    widget.account, widget.history_limit
                )
                self.application.transfers.network_fetch_total_count_for_address(
                    widget.account
                )

    def fetch_data_from_network_datapod(self):
        """
        Fetch data from the network data pod

        :return:
        """
        self.application.datapods.network_fetch_current_datapod()

        for account in self.application.accounts.get_list():
            try:
                self.application.profiles.network_update(account)
            except Exception as exception:
                logging.exception(exception)

    def connect_to_servers_and_fetch_data_finished(self):
        """
        Triggered when all connections are done

        :return:
        """
        self.event_bus.repository_data_updated.emit()
        logging.debug("Network servers data fetch finished....emit signal")

    #########################################
    # EVENT BUS HANDLING
    #########################################
    def on_window_shown(self):
        """
        Triggered when the window is shown

        :return:
        """
        self.thread_manager.start(
            self.connect_to_servers_and_fetch_data,
            self.connect_to_servers_and_fetch_data_finished,
        )

    def on_user_data_restored_event(self):
        """
        Triggered when the user data are restored

        :return:
        """
        self.thread_manager.start(
            self.connect_to_servers_and_fetch_data,
            self.connect_to_servers_and_fetch_data_finished,
        )

    def on_currency_changed(self, current_code_name: str, new_code_name: str):
        """
        Triggered when the currency_changed event is dispatched

        :param current_code_name: Current Currency code name
        :param new_code_name: New Currency code name
        :return:
        """
        # disconnect servers
        self.application.connections.disconnect_all()

        # close current currency
        self.event_bus.currency_closed.emit(current_code_name)

        # change currency in backend
        self.application.select_currency(new_code_name)

        # close current currency
        self.event_bus.currency_opened.emit(new_code_name)

        # async update data from network
        self.thread_manager.start(
            self.connect_to_servers_and_fetch_data,
            self.connect_to_servers_and_fetch_data_finished,
        )

    def on_transfer_sent(self, accounts: List[Account]):
        """
        Triggered when a transfer_sent event is dispatched

        :param accounts: List of Account instances
        :return:
        """
        if len(accounts) > 0:
            logging.debug(
                "QMainApplication.on_transfer_sent triggered !!!!!!!!!!!!!!!!!!!!!"
            )
            self.application.accounts.network_update_balances(accounts),

            for account in accounts:
                self.application.transfers.network_fetch_history_for_account(
                    account, ACCOUNT_TRANSFER_HISTORY_LIMIT
                )
                self.application.accounts.network_update_total_transfers_count(account)

    def on_transfer_sent_finished(self, accounts: List[Account]):
        """
        Executed after a transfer_sent event is dispatched

        :param accounts: List of Account instances
        :return:
        """
        self.event_bus.account_balances_updated.emit(accounts)
        logging.debug(
            "QMainApplication.on_transfer_sent_finished account_balances_updated emitted !!!!!!!!!!!!!!!!!!!!!"
        )

    def on_identity_created(self, issuer_identity: Identity, identity: Identity):
        """
        Triggered when an identity_created event is dispatched

        :param issuer_identity: Issuer Identity instance
        :param identity: Created Identity instance
        :return:
        """
        # wait for a block before getting indexed data
        sleep(self.application.currencies.get_current().block_duration / 1000)

        try:
            self.application.certifications.network_fetch_certifications_for_identity(
                issuer_identity.index
            )
        except Exception as exception:
            logging.exception(exception)

        try:
            self.application.certifications.network_fetch_certifications_for_identity(
                identity.index
            )
        except Exception as exception:
            logging.exception(exception)

    def on_identity_created_finished(
        self, issuer_identity: Identity, identity: Identity
    ):
        """
        Executed after an identity_created event is dispatched

        :param issuer_identity: Issuer Identity instance
        :param identity: Created Identity instance
        :return:
        """
        self.event_bus.identities_updated.emit([issuer_identity, identity])
        logging.debug(
            "QMainApplication.on_identity_created_finished identities_updated emitted !!!!!!!!!!!!!!!!!!!!!"
        )

    def on_identity_confirmed(self, identity: Identity):
        """
        Triggered when an identity_confirmed event is dispatched

        :param identity: Confirmed Identity instance
        :return:
        """
        # wait for a block before getting indexed data
        sleep(self.application.currencies.get_current().block_duration / 1000)

        # update identity again to get indexed name
        try:
            self.application.identities.network_update_identity(identity.address)
        except Exception as exception:
            logging.exception(exception)
        # update certifications with name
        try:
            self.application.certifications.network_fetch_certifications_for_identity(
                identity.index
            )
        except Exception as exception:
            logging.exception(exception)

        certifications = self.application.certifications.list(identity.index)
        if len(certifications) > 0:
            # update first certification of identity creator
            try:
                self.application.certifications.network_fetch_certifications_for_identity(
                    certifications[0].issuer_identity_index
                )
            except Exception as exception:
                logging.exception(exception)

    def on_identity_confirmed_finished(self, identity: Identity):
        """
        Executed after an identity_created event is dispatched

            :param identity: Confirmed Identity instance
        :return:
        """
        identities = [identity]
        certifications = self.application.certifications.list(identity.index)
        if len(certifications) > 0:
            identity_creator = self.application.identities.get_by_address(
                certifications[0].issuer_address
            )
            if identity_creator is not None:
                identities.append(identity_creator)

        self.event_bus.identities_updated.emit(identities)
        logging.debug(
            "QMainApplication.on_identity_confirmed_finished identities_updated emitted !!!!!!!!!!!!!!!!!!!!!"
        )

    def on_identity_certified(
        self, issuer_identity: Identity, receiver_identity: Identity
    ):
        """
        Triggered when an identity_certified event is dispatched

        :param issuer_identity: Issuer Identity instance
        :param receiver_identity: Receiver Identity instance
        :return:
        """
        # wait for a block before getting indexed data
        sleep(self.application.currencies.get_current().block_duration / 1000)

        # update receiver identity if status changed
        try:
            self.application.identities.network_update_identity(
                receiver_identity.address
            )
        except Exception as exception:
            logging.exception(exception)

        try:
            self.application.certifications.network_fetch_certifications_for_identity(
                issuer_identity.index
            )
        except Exception as exception:
            logging.exception(exception)

        try:
            self.application.certifications.network_fetch_certifications_for_identity(
                receiver_identity.index
            )
        except Exception as exception:
            logging.exception(exception)

    def on_identity_certified_finished(
        self, issuer_identity: Identity, receiver_identity: Identity
    ):
        """
        Executed after an identity_created event is dispatched

        :param issuer_identity: Issuer Identity instance
        :param receiver_identity: Receiver Identity instance
        :return:
        """
        self.event_bus.identities_updated.emit([issuer_identity, receiver_identity])
        logging.debug(
            "QMainApplication.on_identity_certified_finished identities_updated emitted !!!!!!!!!!!!!!!!!!!!!"
        )

    def on_identity_account_changed(
        self, issuer_account: Account, destination_account: Account
    ):
        """
        Triggered when an identity_account_changed event is dispatched

        :param issuer_account: Issuer Account instance
        :param destination_account: Destination Account instance
        :return:
        """
        # wait for a block before getting indexed data
        sleep(self.application.currencies.get_current().block_duration / 1000)

        # update issuer identity (should be None)
        try:
            self.application.identities.network_update_identity(issuer_account.address)
        except Exception as exception:
            logging.exception(exception)

        # update destination identity
        try:
            self.application.identities.network_update_identity(
                destination_account.address
            )
        except Exception as exception:
            logging.exception(exception)

    def on_identity_account_changed_finished(
        self, issuer_account: Account, destination_account: Account
    ):
        """
        Triggered when an identity_account_changed event is finished

        :param issuer_account: Issuer Account instance
        :param destination_account: Destination Account instance
        :return:
        """
        self.event_bus.account_identities_updated.emit(
            [issuer_account, destination_account]
        )
        logging.debug(
            "QMainApplication.on_identity_account_changed_finished account_identities_updated emitted !!!!!!!!!!!!!!!!!!!!!"
        )

    def on_profile_account_changed(
        self, issuer_account: Account, destination_account: Account
    ):
        """
        Triggered when a profile_account_changed event is dispatched

        :param issuer_account: Issuer Account instance
        :param destination_account: Destination Account instance
        :return:
        """
        pass

    def on_profile_account_changed_finished(
        self, issuer_account: Account, destination_account: Account
    ):
        """
        Triggered when on_profile_account_changed event is finished

        :param issuer_account: Issuer Account instance
        :param destination_account: Destination Account instance
        :return:
        """
        self.event_bus.account_profiles_updated.emit(
            [issuer_account, destination_account]
        )
        logging.debug(
            "QMainApplication.on_profile_account_changed_finished account_profiles_updated emitted !!!!!!!!!!!!!!!!!!!!!"
        )



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
from typing import Any, Dict, List, Optional

from tikka.domains.entities.identity import Certification


class CertificationsRepositoryInterface(abc.ABC):
    """
    CertificationsRepositoryInterface class
    """

    COLUMN_ID = "certification_id"
    COLUMN_ISSUER_IDENTITY_INDEX = "certification_issuer_identity_index"
    COLUMN_ISSUER_IDENTITY_NAME = "certification_issuer_identity_name"
    COLUMN_ISSUER_IDENTITY_STATUS = "certification_issuer_identity_status"
    COLUMN_ISSUER_ADDRESS = "certification_issuer_address"
    COLUMN_RECEIVER_IDENTITY_INDEX = "certification_receiver_identity_index"
    COLUMN_RECEIVER_IDENTITY_NAME = "certification_receiver_identity_name"
    COLUMN_RECEIVER_IDENTITY_STATUS = "certification_receiver_identity_status"
    COLUMN_RECEIVER_ADDRESS = "certification_receiver_address"
    COLUMN_CREATED_ON = "certification_created_on"
    COLUMN_UPDATED_ON = "certification_updated_on"
    COLUMN_EXPIRE_ON = "certification_expire_on"

    SORT_ORDER_ASCENDING = "ASC"
    SORT_ORDER_DESCENDING = "DESC"

    @abc.abstractmethod
    def set_list(
        self, identity_index: int, certifications: List[Certification]
    ) -> None:
        """
        Set identity certification list in repository

        :param identity_index: Identity index
        :param certifications: List of Certification instances
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(
        self,
        identity_index: int,
        filters: Optional[Dict[str, Any]] = None,
        sort_column: str = COLUMN_CREATED_ON,
        sort_order: str = SORT_ORDER_ASCENDING,
        limit: int = 10,
        offset: int = 0,
    ) -> List[Certification]:
        """
        List certifications from and to identity index, from repository with optional filters and sort_column

        :param identity_index: Identity index
        :param filters: Dict with {column: value} filters or None
        :param sort_column: Sort column constant, default to COLUMN_TIMESTAMP
        :param sort_order: Sort order constant SORT_ORDER_ASCENDING or SORT_ORDER_DESCENDING
        :param limit: Number of rows to return
        :param offset: Offset of first row to return
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_by_address(self, address: str, exclude: List[str]) -> None:
        """
        Delete certification history issued from or received by address
        (exclude certifications with address in exclude list)

        :param address: Account address of certifications to delete
        :param exclude: List of address to not delete
        :return:
        """
        raise NotImplementedError

    @abc.abstractmethod
    def delete_all(self) -> None:
        """
        Delete all certifications in repository

        :return:
        """
        raise NotImplementedError



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'tikka/slots/pyqt/resources/gui/windows/profile_change_account.ui'
#
# Created by: PyQt5 UI code generator 5.15.10
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_ProfileChangeAccountDialog(object):
    def setupUi(self, ProfileChangeAccountDialog):
        ProfileChangeAccountDialog.setObjectName("ProfileChangeAccountDialog")
        ProfileChangeAccountDialog.resize(738, 172)
        self.verticalLayout = QtWidgets.QVBoxLayout(ProfileChangeAccountDialog)
        self.verticalLayout.setObjectName("verticalLayout")
        self.helpLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        font = QtGui.QFont()
        font.setItalic(True)
        self.helpLabel.setFont(font)
        self.helpLabel.setObjectName("helpLabel")
        self.verticalLayout.addWidget(self.helpLabel)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.addressLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressLabel.sizePolicy().hasHeightForWidth())
        self.addressLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.addressLabel.setFont(font)
        self.addressLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.addressLabel.setObjectName("addressLabel")
        self.gridLayout.addWidget(self.addressLabel, 1, 0, 1, 1)
        self.accountLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountLabel.sizePolicy().hasHeightForWidth())
        self.accountLabel.setSizePolicy(sizePolicy)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.accountLabel.setFont(font)
        self.accountLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.accountLabel.setObjectName("accountLabel")
        self.gridLayout.addWidget(self.accountLabel, 0, 0, 1, 1)
        self.profileLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.profileLabel.setFont(font)
        self.profileLabel.setAlignment(QtCore.Qt.AlignRight|QtCore.Qt.AlignTrailing|QtCore.Qt.AlignVCenter)
        self.profileLabel.setObjectName("profileLabel")
        self.gridLayout.addWidget(self.profileLabel, 2, 0, 1, 1)
        self.profileValueLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        self.profileValueLabel.setText("")
        self.profileValueLabel.setObjectName("profileValueLabel")
        self.gridLayout.addWidget(self.profileValueLabel, 2, 1, 1, 1)
        self.accountComboBox = QtWidgets.QComboBox(ProfileChangeAccountDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.accountComboBox.sizePolicy().hasHeightForWidth())
        self.accountComboBox.setSizePolicy(sizePolicy)
        self.accountComboBox.setObjectName("accountComboBox")
        self.gridLayout.addWidget(self.accountComboBox, 0, 1, 1, 1)
        self.addressValueLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.addressValueLabel.sizePolicy().hasHeightForWidth())
        self.addressValueLabel.setSizePolicy(sizePolicy)
        self.addressValueLabel.setText("")
        self.addressValueLabel.setObjectName("addressValueLabel")
        self.gridLayout.addWidget(self.addressValueLabel, 1, 1, 1, 1)
        self.changeAccountButton = QtWidgets.QPushButton(ProfileChangeAccountDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Fixed, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.changeAccountButton.sizePolicy().hasHeightForWidth())
        self.changeAccountButton.setSizePolicy(sizePolicy)
        self.changeAccountButton.setObjectName("changeAccountButton")
        self.gridLayout.addWidget(self.changeAccountButton, 0, 2, 1, 1)
        self.loaderIconLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        self.loaderIconLabel.setMaximumSize(QtCore.QSize(16, 16))
        self.loaderIconLabel.setText("")
        self.loaderIconLabel.setScaledContents(True)
        self.loaderIconLabel.setObjectName("loaderIconLabel")
        self.gridLayout.addWidget(self.loaderIconLabel, 0, 3, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.errorLabel = QtWidgets.QLabel(ProfileChangeAccountDialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.errorLabel.sizePolicy().hasHeightForWidth())
        self.errorLabel.setSizePolicy(sizePolicy)
        self.errorLabel.setStyleSheet("color: red")
        self.errorLabel.setText("")
        self.errorLabel.setAlignment(QtCore.Qt.AlignCenter)
        self.errorLabel.setObjectName("errorLabel")
        self.verticalLayout.addWidget(self.errorLabel)
        self.buttonBox = QtWidgets.QDialogButtonBox(ProfileChangeAccountDialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Close)
        self.buttonBox.setCenterButtons(True)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)

        self.retranslateUi(ProfileChangeAccountDialog)
        self.buttonBox.accepted.connect(ProfileChangeAccountDialog.accept) # type: ignore
        self.buttonBox.rejected.connect(ProfileChangeAccountDialog.reject) # type: ignore
        QtCore.QMetaObject.connectSlotsByName(ProfileChangeAccountDialog)

    def retranslateUi(self, ProfileChangeAccountDialog):
        #  _translate = QtCore.QCoreApplication.translate
        ProfileChangeAccountDialog.setWindowTitle(_("Change account of profile"))
        self.helpLabel.setText(_("Select new account for profile"))
        self.addressLabel.setText(_("Address"))
        self.accountLabel.setText(_("Account"))
        self.profileLabel.setText(_("Profile"))
        self.changeAccountButton.setText(_("Change account"))
import tikka.slots.pyqt.resources.icons.index_rc



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from typing import Any, Dict


def create_profile_html(
    profile_data: Dict[str, Any], updated_on_localized_string: str
) -> str:
    """
    Return HTML content string of Cesium+ Profile

    :param profile_data:
    :param updated_on_localized_string: Updated date localized string
    :return:
    """
    # display avatar image
    avatar_html = ""
    if "avatar" in profile_data and "_content" in profile_data["avatar"]:
        mime_type = profile_data["avatar"].get("_content_type", "image/png")
        avatar_data = profile_data["avatar"]["_content"]
        avatar_html = f'<img src="data:{mime_type};base64,{avatar_data}" alt="Avatar" class="avatar">'

    # web site button
    web_button_html = ""
    if "socials" in profile_data:
        for social in profile_data["socials"]:
            if social.get("type") == "web":
                web_url = social.get("url", "#")
                web_button_html = f'<a href="{web_url}" class="web-button" onclick="return true;">🌐 {web_url}</a>'

    # extract geolocation data
    geo_point = profile_data.get("geoPoint", {}) or {}
    lat = geo_point.get("lat", 0)
    lon = geo_point.get("lon", 0)
    city = profile_data.get("city", "Localisation inconnue")

    html_content = f"""
    <!DOCTYPE html>
    <html>
    <head>
        <meta charset="UTF-8">
        <link rel="stylesheet" href="https://unpkg.com/leaflet@1.9.4/dist/leaflet.css" />
        <script src="https://unpkg.com/leaflet@1.9.4/dist/leaflet.js"></script>
        <style>
            {get_css()}
        </style>
    </head>
    <body>
        <div class="profile-container">
            <div class="profile-header">
                <div class="profile-main">
                    {avatar_html}
                    <div class="profile-info">
                        <h1 class="profile-title">{profile_data.get('title', '')}</h1>
                        <p class="profile-description">{profile_data.get('description', '')}</p>
                        <div class="profile-location">
                            📍 {city}
                        </div>
                        {web_button_html}
                        <div class="profile-date">{updated_on_localized_string}</div>
                    </div>
                </div>
            </div>

            <div class="map-column">
                <div id="minimap" data-lat="{lat}" data-lon="{lon}" data-city="{city}"></div>
            </div>
        </div>

        <script>
            {get_map_js()}
        </script>
    </body>
    </html>
    """

    return html_content


def get_css() -> str:
    """
    Return Profile CSS

    :return:
    """
    return """
    * {
        margin: 0;
        padding: 0;
        box-sizing: border-box;
    }

    body {
        font-family: -apple-system, BlinkMacSystemFont, 'Segoe UI', Roboto, sans-serif;
        /* background: linear-gradient(135deg, #667eea 0%, #764ba2 100%);*/
        min-height: 100vh;
        padding: 20px;
        display: flex;
        justify-content: center;
        align-items: flex-start;
    }

    .profile-container {
        display: flex;
        gap: 20px;
        /* max-width: 1000px;*/
        width: 100%;
    }

    .profile-header {
        background: linear-gradient(135deg, #4facfe 0%, #00f2fe 100%);
        border-radius: 12px;
        box-shadow: 0 10px 30px rgba(0,0,0,0.2);
        padding: 30px;
        flex: 1;
        color: white;
        display: flex;
        flex-direction: column;
        justify-content: space-between;
    }

    .profile-main {
        display: flex;
        align-items: flex-start;
        gap: 20px;
        margin-bottom: 20px;
    }

    .avatar {
        width: 80px;
        height: 80px;
        border-radius: 50%;
        border: 3px solid white;
        object-fit: cover;
        flex-shrink: 0;
    }

    .profile-info {
        flex: 1;
    }

    .profile-title {
        font-size: 24px;
        font-weight: 600;
        margin-bottom: 8px;
    }

.profile-description {
    font-size: 16px;
    opacity: 0.9;
    margin-bottom: 12px;
    line-height: 1.4;
    white-space: pre-wrap;      /* Respecte les sauts de ligne */
    word-wrap: break-word;      /* Casse les mots longs */
    overflow-wrap: break-word;  /* Alternative moderne */
}

    .profile-location {
        font-size: 14px;
        opacity: 0.9;
        margin-bottom: 15px;
    }

    .web-button {
        display: inline-block;
        padding: 10px 20px;
        background: rgba(255, 255, 255, 0.2);
        color: white;
        text-decoration: none;
        border-radius: 25px;
        border: 2px solid white;
        font-weight: 600;
        transition: all 0.3s ease;
        margin-bottom: 10px;
    }

    .web-button:hover {
        background: white;
        color: #4facfe;
        transform: translateY(-2px);
    }

    .profile-date {
        font-size: 12px;
        opacity: 0.8;
        font-style: italic;
    }

    .map-column {
        flex: 0 0 400px;
    }

    #minimap {
        height: 100%;
        min-height: 300px;
        border-radius: 12px;
        box-shadow: 0 10px 30px rgba(0,0,0,0.2);
        overflow: hidden;
        background: white;
    }

    /* Style pour la carte Leaflet */
    .leaflet-container {
        background: #f8f9fa;
        border-radius: 12px;
    }

    .custom-marker {
        background: #e74c3c;
        border: 3px solid white;
        border-radius: 50%;
        width: 20px;
        height: 20px;
        box-shadow: 0 2px 10px rgba(0,0,0,0.3);
    }

    /* Responsive pour petits écrans */
    @media (max-width: 768px) {
        .profile-container {
            flex-direction: column;
        }

        .map-column {
            flex: 0 0 250px;
        }
    }
    """


def get_map_js() -> str:
    """
    Return Profile Javascript

    :return:
    """
    return """
    function initMap() {
        const mapElement = document.getElementById('minimap');
        const lat = parseFloat(mapElement.dataset.lat);
        const lon = parseFloat(mapElement.dataset.lon);
        const city = mapElement.dataset.city;

        // Vérifier si les coordonnées sont valides
        if (lat === 0 && lon === 0) {
            mapElement.innerHTML = '<div style="display: flex; align-items: center; justify-content: center; height: 100%; background: #f8f9fa; color: #6c757d; font-size: 16px;">📍 Localisation non disponible</div>';
            return;
        }

        // Initialiser la carte
        const map = L.map('minimap').setView([lat, lon], 13);

        // Ajouter les tuiles OpenStreetMap
        L.tileLayer('https://{s}.tile.openstreetmap.org/{z}/{x}/{y}.png', {
            attribution: '© OpenStreetMap contributors',
            maxZoom: 18
        }).addTo(map);

        // Créer un marqueur personnalisé
        const customIcon = L.divIcon({
            className: 'custom-marker',
            html: '📍',
            iconSize: [30, 30],
            iconAnchor: [15, 30]
        });

        // Ajouter le marqueur
        const marker = L.marker([lat, lon], {icon: customIcon})
            .addTo(map)
            .bindPopup(`<b>${city}</b><br>${lat.toFixed(6)}, ${lon.toFixed(6)}`)
            .openPopup();
    }

    // Initialiser la carte quand la page est chargée
    document.addEventListener('DOMContentLoaded', initMap);
    """



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import logging
import sys
from datetime import datetime
from typing import List, Optional

import qrcode
from PyQt5.QtCore import QEvent, QLocale, QPoint, QTimer, pyqtSignal
from PyQt5.QtGui import QDesktopServices, QFont, QPixmap, QWheelEvent
from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.QtWidgets import QApplication, QListView, QMainWindow, QSizePolicy, QWidget

from tikka.domains.application import Application
from tikka.domains.entities.account import Account, AccountCryptoType
from tikka.domains.entities.address import DisplayAddress
from tikka.domains.entities.constants import AMOUNT_UNIT_KEY, DATA_PATH
from tikka.domains.entities.events import (
    UnitEvent,
)
from tikka.domains.entities.identity import Identity, IdentityStatus
from tikka.domains.entities.profile import Profile
from tikka.domains.entities.smith import SmithStatus
from tikka.slots.pyqt.entities.constants import (
    ACCOUNT_TRANSFER_HISTORY_LIMIT,
    ADDRESS_MONOSPACE_FONT_NAME,
    ICON_ACCOUNT_NO_WALLET,
    ICON_ACCOUNT_WALLET_LOCKED,
    ICON_ACCOUNT_WALLET_UNLOCKED,
    ICON_IDENTITY_MEMBER,
    ICON_IDENTITY_MEMBER_NOT_VALIDATED,
    ICON_IDENTITY_NOT_MEMBER,
    SELECTED_UNIT_PREFERENCES_KEY,
)
from tikka.slots.pyqt.entities.event_bus import EventBus
from tikka.slots.pyqt.entities.qrcode_image import QRCodeImage
from tikka.slots.pyqt.entities.worker import ThreadManager
from tikka.slots.pyqt.models.certifications import (
    CertificationItemDelegate,
    CertificationsListModel,
)
from tikka.slots.pyqt.models.transfers import TransferItemDelegate, TransfersListModel
from tikka.slots.pyqt.resources.gui.widgets.account_rc import Ui_AccountWidget
from tikka.slots.pyqt.widgets.account_certifications_menu import (
    AccountCertificationsPopupMenu,
)
from tikka.slots.pyqt.widgets.account_menu import AccountPopupMenu
from tikka.slots.pyqt.widgets.account_profile_view import create_profile_html
from tikka.slots.pyqt.widgets.account_transfers_menu import AccountTransfersPopupMenu
from tikka.slots.pyqt.windows.identity_certify import IdentityCertifyWindow
from tikka.slots.pyqt.windows.identity_change_account import IdentityChangeAccountWindow
from tikka.slots.pyqt.windows.identity_confirm import IdentityConfirmWindow
from tikka.slots.pyqt.windows.identity_create import IdentityCreateWindow
from tikka.slots.pyqt.windows.profile_change_account import ProfileChangeAccountWindow
from tikka.slots.pyqt.windows.transfer import TransferWindow


class TransfersListView(QListView):
    """
    TransfersListView class
    """

    scroll_offset_changed = pyqtSignal(int)

    def __init__(
        self,
        application: Application,
        address: str,
        page_size: int,
        parent: Optional[QWidget] = None,
    ):
        """
        Init TransfersListView instance

        :param application: Application instance
        :param address: Account address
        :param page_size: Nb of transfers per page
        :param parent: QWidget instance, default None
        """
        super().__init__(parent)

        self.application = application
        self.address = address
        self.scroll_offset = 0
        self.page_size = page_size

    def wheelEvent(self, event: QWheelEvent):
        """
        Override Wheel event

        :param event: QEvent instance
        :return:
        """
        delta = event.angleDelta().y()
        if not self.verticalScrollBar().isVisible():
            if delta < 0:  # 🔽 Scroll vers le bas (charger plus)
                if (
                    self.scroll_offset + self.page_size
                    < self.application.transfers.count(self.address)
                ):
                    self.scroll_offset += self.page_size
                    self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                    self.scroll_offset_changed.emit(self.scroll_offset)

            elif (
                delta > 0
            ):  # 🔼 Scroll vers le haut (charger page précédente si disponible)
                self.scroll_offset -= self.page_size
                if self.scroll_offset < 0:
                    self.scroll_offset = 0
                self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                self.scroll_offset_changed.emit(self.scroll_offset)
        else:
            # wheel backward and scroll bar maximum down...
            if (
                delta < 0
                and self.verticalScrollBar().value()
                == self.verticalScrollBar().maximum()
            ):

                if (
                    self.scroll_offset + self.page_size
                    < self.application.transfers.count(self.address)
                ):
                    self.scroll_offset += self.page_size
                    self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                    self.scroll_offset_changed.emit(self.scroll_offset)

            # wheel forward and scroll bar maximum top...
            elif (
                delta > 0
                and self.verticalScrollBar().value()
                == self.verticalScrollBar().minimum()
            ):
                self.scroll_offset -= self.page_size
                if self.scroll_offset < 0:
                    self.scroll_offset = 0
                self.model().init_data(self.scroll_offset, self.page_size)  # type: ignore
                self.scroll_offset_changed.emit(self.scroll_offset)

        super().wheelEvent(event)


class AccountWidget(QWidget, Ui_AccountWidget):
    """
    AccountWidget class
    """

    def __init__(
        self,
        application: Application,
        account: Account,
        parent: Optional[QWidget] = None,
    ) -> None:
        """
        Init AccountWidget instance

        :param application: Application instance
        :param account: Account instance
        :param parent: QWidget instance
        """
        super().__init__(parent=parent)
        self.setupUi(self)

        self.application = application
        self._ = self.application.translator.gettext

        self.event_bus = EventBus.instance()
        self.event_bus.repository_data_updated.connect(
            self.repository_data_updated_event_bus
        )
        self.event_bus.account_renamed.connect(self.on_account_renamed_event_bus)
        self.event_bus.account_wallet_updated.connect(
            self.on_account_wallet_updated_event_bus
        )
        self.event_bus.account_balances_updated.connect(
            self.on_account_balances_updated
        )
        self.event_bus.identities_updated.connect(self.on_identities_updated_event_bus)
        self.event_bus.account_identities_updated.connect(
            self.on_account_identities_updated_event_bus
        )
        self.event_bus.account_profiles_updated.connect(
            self.on_account_profiles_updated_event_bus
        )

        self.account: Account = account
        self.identity: Optional[Identity] = self.application.identities.get_by_address(
            self.account.address
        )

        self.history_limit = ACCOUNT_TRANSFER_HISTORY_LIMIT
        self.history_page_size = 100
        self.history_scroll_offset = 0
        self.profile: Optional[Profile] = self.application.profiles.get(
            self.account.address
        )

        self.monospace_font = QFont(ADDRESS_MONOSPACE_FONT_NAME)
        self.monospace_font.setStyleHint(QFont.Monospace)
        self.addressValueLabel.setFont(self.monospace_font)
        self.addressValueLabel.setText(self.account.address)
        if self.account.legacy_v1:
            self.v1AddressValueLabel.setText(
                self.account.get_v1_address(
                    self.application.currencies.get_current().ss58_format, True
                )
            )

        # creating a pix map of qr code
        qr_code_pixmap = qrcode.make(
            self.account.address, image_factory=QRCodeImage
        ).pixmap()

        # set qrcode to the label
        self.QRCodeAddressLabel.setPixmap(qr_code_pixmap)

        self.transfers_model = TransfersListModel(
            application, self.account.address, self.locale()
        )

        # self.transfersListView.hide()
        # self.transfersListView = TransfersListView(
        #     application, self.account.address, self.history_page_size, self.groupBox
        # )
        #
        # self.transfersListView.setMouseTracking(False)
        # self.transfersListView.setEditTriggers(QAbstractItemView.NoEditTriggers)
        # self.transfersListView.setAlternatingRowColors(True)
        # self.transfersListView.setSelectionMode(QAbstractItemView.NoSelection)
        # self.transfersListView.setObjectName("transfersListView")
        # self.verticalLayout_4.addWidget(self.transfersListView)

        # transfers
        self.transfersListView.setModel(self.transfers_model)
        self.transfersListView.setItemDelegate(TransferItemDelegate())
        self.transfersListView.setResizeMode(QListView.Adjust)

        # certifications received
        self.certifications_received_model = CertificationsListModel(
            application,
            self.locale(),
            self.application.repository.certifications.COLUMN_RECEIVER_IDENTITY_INDEX,
        )
        self.certificationsReceivedListView.setModel(self.certifications_received_model)
        self.certificationsReceivedListView.setItemDelegate(CertificationItemDelegate())
        self.certificationsReceivedListView.setResizeMode(QListView.Adjust)

        # certifications issued
        self.certifications_issued_model = CertificationsListModel(
            application,
            self.locale(),
            self.application.repository.certifications.COLUMN_ISSUER_IDENTITY_INDEX,
        )
        self.certificationsIssuedListView.setModel(self.certifications_issued_model)
        self.certificationsIssuedListView.setItemDelegate(CertificationItemDelegate())
        self.certificationsIssuedListView.setResizeMode(QListView.Adjust)

        self.certifications_received_model.init_data(
            self.identity.index if self.identity else None
        )
        self.certifications_issued_model.init_data(
            self.identity.index if self.identity else None
        )

        # profile webview
        self.profile_web_view = QWebEngineView(self.tabWidgetPage3)
        self.profileWebViewFrame.layout().addWidget(self.profile_web_view)
        self.profile_web_view.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Expanding
        )
        # Intercepter les changements d'URL pour ouvrir les liens externes
        self.profile_web_view.urlChanged.connect(self.on_profile_web_view_url_changed)

        ##############################
        # ASYNC METHODS
        ##############################
        self.thread_manager: ThreadManager = ThreadManager(self)

        # events
        self.refreshButton.clicked.connect(self._on_refresh_button_clicked)
        self.transferToButton.clicked.connect(self.transfer_to)
        self.transferFromButton.clicked.connect(self.transfer_from)
        self.customContextMenuRequested.connect(self.on_account_context_menu)
        # self.transfersListView.scroll_offset_changed.connect(self._update_ui)
        self.transfersListView.customContextMenuRequested.connect(
            self.on_transfers_context_menu
        )
        self.confirmIdentityButton.clicked.connect(
            self.on_confirm_identity_button_clicked
        )
        self.createIdentityButton.clicked.connect(
            self.on_create_identity_button_clicked
        )
        self.certifyIdentityButton.clicked.connect(
            self.on_certify_identity_button_clicked
        )
        self.identityChangeAccountButton.clicked.connect(
            self.on_identity_change_account_button_clicked
        )
        self.certificationsReceivedListView.customContextMenuRequested.connect(
            self.on_received_certifications_context_menu
        )
        self.certificationsIssuedListView.customContextMenuRequested.connect(
            self.on_issued_certifications_context_menu
        )
        self.profileChangeAccountButton.clicked.connect(
            self.on_profile_change_account_button_clicked
        )
        # application events
        self.application.event_dispatcher.add_event_listener(
            UnitEvent.EVENT_TYPE_CHANGED, lambda e: self._on_unit_changed_event()
        )

        self.transfers_model.init_data(
            self.history_scroll_offset, self.history_page_size
        )

        self._update_ui()

    def cleanup(self):
        """
        Clean up WebEngine resources

        Fix message: Release of profile requested but WebEnginePage still not deleted. Expect troubles !

        :return:
        """
        if hasattr(self, "profile_web_view"):
            # Clear the profile web view
            self.profile_web_view.deleteLater()
            del self.profile_web_view

    def closeEvent(self, event):
        """
        Triggered when user close main window or tab

        :param event:
        :return:
        """
        self.cleanup()
        super().closeEvent(event)

    def transfer_to(self):
        """
        When user click on transfer to button

        :return:
        """
        TransferWindow(
            self.application, None, self.account, parent=self.parent()
        ).exec_()

    def transfer_from(self):
        """
        When user click on transfer from button

        :return:
        """
        TransferWindow(
            self.application, self.account, None, parent=self.parent()
        ).exec_()

    def _on_refresh_button_clicked(self, _: QEvent):
        """
        Triggered when user click on Refresh button

        :param _: QEvent instance
        :return:
        """
        self.errorLabel.setText("")
        self.thread_manager.start(
            self.fetch_account_from_network,
            self._on_finished_fetch_account_from_network,
        )
        self.thread_manager.start(
            self.fetch_transfers_from_network,
            self.on_finished_fetch_transfers_from_network,
        )
        self.thread_manager.start(
            self.fetch_certifications_from_network,
            self.on_finished_fetch_certifications_from_network,
        )
        self.thread_manager.start(
            self.fetch_profile_from_network,
            self._on_finished_fetch_profile_from_network,
        )

    def _on_unit_changed_event(self):
        """
        Triggered when user change money units

        :return:
        """
        self._update_ui()
        self.transfers_model.init_data(
            self.history_scroll_offset, self.history_page_size
        )

    def fetch_account_from_network(self):
        """
        Get last account data from the network

        :return:
        """
        if not self.application.connections.node.is_connected():
            return

        self.refreshButton.setEnabled(False)

        try:
            self.application.accounts.network_update_account(self.account)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)
        else:
            try:
                identity = self.application.identities.network_update_identity(
                    self.account.address
                )
            except Exception as exception:
                self.errorLabel.setText(self._(str(exception)))
                logging.exception(exception)
            else:
                if identity is not None:
                    try:
                        self.application.smiths.network_update_smith(identity.index)
                    except Exception as exception:
                        self.errorLabel.setText(self._(str(exception)))
                        logging.exception(exception)

    def _on_finished_fetch_account_from_network(self):
        """
        Triggered when async request fetch_account_from_network is finished

        :return:
        """
        self._update_ui()
        self.refreshButton.setEnabled(True)
        logging.debug("Account widget update %s", self.account.address)

    def fetch_transfers_from_network(self):
        """
        Get transfers data from the network

        :return:
        """
        if not self.application.connections.indexer.is_connected():
            return

        self.refreshButton.setEnabled(False)

        try:
            self.application.transfers.network_fetch_history_for_account(
                self.account, self.history_limit
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

        try:
            self.application.accounts.network_update_total_transfers_count(self.account)
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def on_finished_fetch_transfers_from_network(self):
        """
        Triggered when async request fetch_transfers_from_network is finished

        :return:
        """
        self.account = self.application.accounts.get_by_address(self.account.address)
        self.transfers_model.init_data(
            self.history_scroll_offset, self.history_page_size
        )
        self._update_ui()
        self.refreshButton.setEnabled(True)
        logging.debug(
            "Finished fetch account transfers %s %s",
            self.account.name,
            self.account.address,
        )

    def fetch_certifications_from_network(self):
        """
        Get certifications data from the network

        :return:
        """
        if (
            self.identity is None
            or not self.application.connections.indexer.is_connected()
        ):
            return

        self.refreshButton.setEnabled(False)

        try:
            self.application.certifications.network_fetch_certifications_for_identity(
                self.identity.index
            )
        except Exception as exception:
            self.errorLabel.setText(self._(str(exception)))
            logging.exception(exception)

    def on_finished_fetch_certifications_from_network(self):
        """
        Triggered when async request fetch_certifications_from_network is finished

        :return:
        """
        self.certifications_received_model.init_data(
            self.identity.index if self.identity else None
        )
        self.certifications_issued_model.init_data(
            self.identity.index if self.identity else None
        )

        self.refreshButton.setEnabled(True)
        logging.debug(
            "Finished fetch identity certifications %s %s",
            self.account.name,
            self.account.address,
        )

    def fetch_profile_from_network(self):
        """
        Get last account data from the network

        :return:
        """
        if not self.application.connections.datapod.is_connected():
            return

        self.refreshButton.setEnabled(False)
        try:
            self.profile = self.application.profiles.network_update(self.account)
        except Exception as exception:
            self.errorLabel.setText(str(exception))
            logging.exception(exception)

    def _on_finished_fetch_profile_from_network(self):
        """
        Triggered when async request fetch_account_from_network is finished

        :return:
        """
        self._update_ui()
        self.refreshButton.setEnabled(True)
        logging.debug("Account profile widget update")

    def _update_ui(self):
        """
        Update UI from self.account

        :return:
        """
        # update local data from DB
        self.account = self.application.accounts.get_by_address(self.account.address)
        self.identity = self.application.identities.get_by_address(self.account.address)
        self.profile = self.application.profiles.get(self.account.address)

        unit_preference = self.application.repository.preferences.get(
            SELECTED_UNIT_PREFERENCES_KEY
        )
        if unit_preference is not None:
            amount = self.application.amounts.get_amount(unit_preference)
        else:
            amount = self.application.amounts.get_amount(AMOUNT_UNIT_KEY)

        if self.account.balance is None:
            self.balanceValueLabel.setText("?")
            self.balanceReservedValueLabel.setText("")
            self.transferFromButton.setEnabled(False)
        else:
            self.balanceValueLabel.setText(
                self.locale().toCurrencyString(
                    amount.value(self.account.balance), amount.symbol()
                )
            )
            display_reserved_balance = self.locale().toCurrencyString(
                amount.value(self.account.balance_reserved), amount.symbol()
            )
            self.balanceReservedValueLabel.setText(f"[-{display_reserved_balance}]")
            self.transferFromButton.setEnabled(True)

        if self.application.wallets.exists(self.account.address):
            if self.account.balance is None or self.account.balance == 0:
                self.transferFromButton.setEnabled(False)
            else:
                self.transferFromButton.setEnabled(True)
            if self.application.wallets.is_unlocked(self.account.address):
                self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_WALLET_UNLOCKED))
            else:
                self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_WALLET_LOCKED))
        else:
            self.lockStatusIcon.setPixmap(QPixmap(ICON_ACCOUNT_NO_WALLET))
            self.transferFromButton.setEnabled(False)

        if self.identity is not None:
            self.accountTabWidget.setTabVisible(1, True)
            display_identity_status = {
                IdentityStatus.UNCONFIRMED.value: self._("Unconfirmed"),
                IdentityStatus.UNVALIDATED.value: self._("Unvalidated"),
                IdentityStatus.MEMBER.value: self._("Member"),
                IdentityStatus.NOT_MEMBER.value: self._("Not member"),
                IdentityStatus.REVOKED.value: self._("Revoked"),
            }

            identity_status_color = {
                IdentityStatus.UNCONFIRMED.value: "orange",
                IdentityStatus.UNVALIDATED.value: "orange",
                IdentityStatus.MEMBER.value: "green",
                IdentityStatus.NOT_MEMBER.value: "red",
                IdentityStatus.REVOKED.value: "red",
            }
            identity_status_style = f"""
                                    color: white;
                                    border: 3px solid "{identity_status_color[self.identity.status.value]}";
                                    border-radius: 10%;
                                    background: "{identity_status_color[self.identity.status.value]}";
                                    """

            display_identity_icon = {
                IdentityStatus.UNCONFIRMED.value: QPixmap(
                    ICON_IDENTITY_MEMBER_NOT_VALIDATED
                ),
                IdentityStatus.UNVALIDATED.value: QPixmap(
                    ICON_IDENTITY_MEMBER_NOT_VALIDATED
                ),
                IdentityStatus.MEMBER.value: QPixmap(ICON_IDENTITY_MEMBER),
                IdentityStatus.NOT_MEMBER.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
                IdentityStatus.REVOKED.value: QPixmap(ICON_IDENTITY_NOT_MEMBER),
            }

            self.identityStatusIconLabel.setPixmap(
                display_identity_icon[self.identity.status.value]
            )
            identity_name = self.identity.name or ""
            self.identityNameValueLabel.setText(
                f"{identity_name}#{self.identity.index}"
            )
            self.identityValueLabel.setText(
                display_identity_status[self.identity.status.value]
            )
            self.identityValueLabel.setStyleSheet(identity_status_style)

            smith = self.application.smiths.get(self.identity.index)
            if smith is not None:
                display_smith_status = {
                    SmithStatus.INVITED.value: self._("Invited"),
                    SmithStatus.PENDING.value: self._("Pending"),
                    SmithStatus.SMITH.value: self._("Smith"),
                    SmithStatus.EXCLUDED.value: self._("Excluded"),
                }
                smith_status_color = {
                    SmithStatus.INVITED.value: "orange",
                    SmithStatus.PENDING.value: "orange",
                    SmithStatus.SMITH.value: "green",
                    SmithStatus.EXCLUDED.value: "red",
                }
                smith_status_style = f"""
                                     color: white;
                                     border: 3px solid "{smith_status_color[smith.status.value]}";
                                     border-radius: 10%;
                                     background: "{smith_status_color[smith.status.value]}";
                                     """
                self.smithValuelabel.setText(display_smith_status[smith.status.value])
                self.smithValuelabel.setStyleSheet(smith_status_style)
            else:
                self.smithValuelabel.setText(self._("No"))
                self.smithValuelabel.setStyleSheet("")

            if self.account.name is None:
                self.accountNameValueLabel.setText("")
            else:
                self.accountNameValueLabel.setText(f" - {self.account.name}")

            self.createIdentityButton.setVisible(
                self.identity.status == IdentityStatus.MEMBER
            )
            self.confirmIdentityButton.setVisible(
                self.identity.status == IdentityStatus.UNCONFIRMED
            )
            self.certifyIdentityButton.setVisible(
                self.identity.status == IdentityStatus.MEMBER
            )
        else:
            self.accountTabWidget.setTabVisible(1, False)
            self.confirmIdentityButton.hide()
            self.createIdentityButton.hide()
            self.certifyIdentityButton.hide()
            self.identityStatusIconLabel.setPixmap(QPixmap(ICON_IDENTITY_NOT_MEMBER))
            self.identityNameValueLabel.setText("")
            self.identityValueLabel.setText(self._("No"))
            self.smithValuelabel.setText(self._("No"))
            self.identityValueLabel.setStyleSheet("")
            self.smithValuelabel.setStyleSheet("")
            if self.account.name is None:
                self.accountNameValueLabel.setText("")
            else:
                self.accountNameValueLabel.setText(self.account.name)

        if self.account.root is None and self.account.path is None:
            self.derivationValueLabel.setText(self._("Root"))
        else:
            root_account = self.application.accounts.get_by_address(self.account.root)
            if root_account is not None and root_account.name is not None:
                self.derivationValueLabel.setFont(QFont())
                self.derivationValueLabel.setText(root_account.name + self.account.path)
            else:
                self.derivationValueLabel.setFont(self.monospace_font)
                self.derivationValueLabel.setText(
                    DisplayAddress(self.account.root).shorten + self.account.path
                )

        # transfers
        self.transfersCountValueLabel.setText(
            f"1...{self.history_limit}/{self.account.total_transfers_count}"
        )
        # current_page = int(
        #     (self.transfersListView.scroll_offset / self.history_page_size) + 1
        # )
        # self.transfersCountValueLabel.setText(
        #     str(
        #         f"Page {current_page}/{self.history_page_size} |"
        #         f"1...{self.history_limit}/{self.account.total_transfers_count}"
        #     )
        # )

        # profile
        if self.profile is not None:
            self.accountTabWidget.setTabVisible(2, True)
            self.profile_web_view.setHtml(self.get_profile_as_html())
        else:
            self.accountTabWidget.setTabVisible(2, False)

        # fix bug on tab titles not showing after setTabVisible()
        QTimer.singleShot(500, self._fix_tab_bar_geometry)

    def _fix_tab_bar_geometry(self):
        """
        Force Qt to calculate again the geometry of tab bar after setTabVisible()
        Fix bug when tab titles do not appear correctly
        """
        try:
            tab_bar = self.accountTabWidget.tabBar()
            if tab_bar:
                current_size = tab_bar.size()

                # Forcer un changement de taille
                tab_bar.resize(current_size.width() + 1, current_size.height())
                QApplication.processEvents()  # Important!
                tab_bar.resize(current_size.width(), current_size.height())
                QApplication.processEvents()
        except RuntimeError:
            # C++ object is destroy, can happen in tests
            pass

    def get_profile_as_html(self) -> str:
        """
        Display profile data as HTML in web view

        :return:
        """
        if self.profile is not None:
            updated_on_time_stamp = self.profile.data.get("time", 0)
            updated_on_localized_datetime_string = self.locale().toString(
                datetime.fromtimestamp(updated_on_time_stamp),
                QLocale.dateTimeFormat(self.locale(), QLocale.ShortFormat),
            )
            updated_on_localized_string = self._(
                "Updated at {localized_string}"
            ).format(localized_string=updated_on_localized_datetime_string)

            return create_profile_html(self.profile.data, updated_on_localized_string)

        return self._("No profile.")

    def on_profile_web_view_url_changed(self, url):
        """
        Triggered when user click on web link in profile web view

        :return:
        """
        if url.isValid() and url.scheme() in ["http", "https"]:
            QDesktopServices.openUrl(url)
            # Reload page to stay on original content
            self.profile_web_view.sender().setHtml(self.get_profile_as_html())

    def on_account_context_menu(self, position: QPoint):
        """
        When right button on account tab

        :return:
        """
        # display popup menu at click position
        menu = AccountPopupMenu(self.application, self.account, self)
        menu.exec_(self.mapToGlobal(position))

    def on_transfers_context_menu(self, position: QPoint):
        """
        When right button on transfers listview

        :return:
        """
        # get selected transfer
        transfer = self.transfersListView.currentIndex().internalPointer()
        if transfer is not None:
            # display popup menu at click position
            menu = AccountTransfersPopupMenu(
                self.application, self.account, transfer, self
            )
            menu.exec_(self.transfersListView.mapToGlobal(position))

    def on_received_certifications_context_menu(self, position: QPoint):
        """
        When right button on received certifications listview

        :return:
        """
        # get selected certification
        certification = (
            self.certificationsReceivedListView.currentIndex().internalPointer()
        )
        if certification is not None:
            # display popup menu at click position
            menu = AccountCertificationsPopupMenu(
                self.application, self.account, certification, self
            )
            menu.exec_(self.certificationsReceivedListView.mapToGlobal(position))

    def on_issued_certifications_context_menu(self, position: QPoint):
        """
        When right button on issued certifications listview

        :return:
        """
        # get selected certification
        certification = (
            self.certificationsIssuedListView.currentIndex().internalPointer()
        )
        if certification is not None:
            # display popup menu at click position
            menu = AccountCertificationsPopupMenu(
                self.application, self.account, certification, self
            )
            menu.exec_(self.certificationsIssuedListView.mapToGlobal(position))

    def on_confirm_identity_button_clicked(self):
        """
        Triggered when user click on confirm_identity_button

        :return:
        """
        window = IdentityConfirmWindow(self.application, self.account)
        window.exec_()

    def on_create_identity_button_clicked(self):
        """
        Triggered when user click on create_identity_button

        :return:
        """
        window = IdentityCreateWindow(self.application, self.account)
        window.exec_()

    def on_certify_identity_button_clicked(self):
        """
        Triggered when user click on certify_identity_button

        :return:
        """
        if self.identity is not None:
            window = IdentityCertifyWindow(self.application, self.account)
            window.exec_()

    def on_identity_change_account_button_clicked(self):
        """
        Triggered when user click on identity change account button

        :return:
        """
        if self.identity is not None:
            window = IdentityChangeAccountWindow(self.application, self.account)
            window.exec_()

    def on_profile_change_account_button_clicked(self):
        """
        Triggered when user click on identity change account button

        :return:
        """
        if self.profile is not None:
            window = ProfileChangeAccountWindow(self.application, self.account)
            window.exec_()

    def on_identities_updated_event_bus(self, identities: List[Identity]):
        """
        Triggered when identities are updated

        :param identities: List of Identity instance
        :return:
        """
        if self.account.address in [identity.address for identity in identities]:
            self._update_ui()
            self.certifications_issued_model.init_data(
                self.identity.index if self.identity else None
            )
            self.certifications_received_model.init_data(
                self.identity.index if self.identity else None
            )

    def on_account_identities_updated_event_bus(self, accounts: List[Account]):
        """
        Triggered when account identities are updated (some identities are None, so we are using accounts!)

        :param accounts: List of Account instance
        :return:
        """
        if self.account.address in [account.address for account in accounts]:
            self._update_ui()
            self.certifications_issued_model.init_data(
                self.identity.index if self.identity else None
            )
            self.certifications_received_model.init_data(
                self.identity.index if self.identity else None
            )

    def on_account_profiles_updated_event_bus(self, accounts: List[Account]):
        """
        Triggered when account profiles are updated (some profiles are None, so we are using accounts!)

        :param accounts: List of Account instance
        :return:
        """
        if self.account.address in [account.address for account in accounts]:
            self._update_ui()

    def on_account_balances_updated(self, accounts: List[Account]):
        """
        Triggered when a on_account_balances_updated event is dispatched

        :param accounts: List of Account instances
        :return:
        """
        if self.account.address in [account.address for account in accounts]:
            self._update_ui()
            self.transfers_model.init_data(
                self.history_scroll_offset, self.history_page_size
            )

    def on_account_wallet_updated_event_bus(self, account: Account):
        """
        Triggered when the account wallet is updated

        :param account: Account instance
        :return:
        """
        self._update_ui()

    def on_account_renamed_event_bus(self, account: Account):
        """
        Triggered when the account is renamed

        :param account: Account instance
        :return:
        """
        self._update_ui()
        self.certifications_issued_model.init_data(
            self.identity.index if self.identity else None
        )
        self.certifications_received_model.init_data(
            self.identity.index if self.identity else None
        )
        self.transfers_model.init_data()

    def repository_data_updated_event_bus(self):
        """
        Triggered when the main window has updated repository data

        :return:
        """
        self._update_ui()
        self.certifications_issued_model.init_data(
            self.identity.index if self.identity else None
        )
        self.certifications_received_model.init_data(
            self.identity.index if self.identity else None
        )
        self.transfers_model.init_data()


if __name__ == "__main__":
    qapp = QApplication(sys.argv)

    application_ = Application(DATA_PATH)
    account_ = application_.accounts.get_by_address(
        "g1LAN1C5rktuWS2giuZ7z1CydYKV2HQeH2rj8A6qY3CemwTVz"
    )
    if account_ is None:
        account_ = application_.accounts.create_new_root_account(
            mnemonic="bottom drive obey lake curtain smoke basket hold race lonely fit walk",
            language_code="en",
            crypto_type=AccountCryptoType.ED25519,
            name="Root ED",
            password="aaaaaa",
        )

    main_window = QMainWindow()
    main_window.show()

    main_window.setCentralWidget(AccountWidget(application_, account_, main_window))

    sys.exit(qapp.exec_())



# Copyright 2021 Vincent Texier <vit@free.fr>
#
# This software is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from dataclasses import dataclass
from typing import Optional

from tikka.interfaces.entities.server import ServerInterface


@dataclass
class Node(ServerInterface):

    url: str
    peer_id: Optional[str] = None
    block: Optional[int] = None
    software: Optional[str] = None
    software_version: Optional[str] = None
    session_keys: Optional[str] = None
    epoch_index: Optional[int] = None
    unsafe_api_exposed: bool = False

    def __str__(self):
        """
        Return string representation

        :return:
        """
        return self.url

    def __eq__(self, other):
        """
        Test equality on url

        :param other: Node instance
        :return:
        """
        if not isinstance(other, self.__class__):
            return False
        return other.url == self.url

    def __hash__(self):
        return hash(self.url)
